from datetime import datetime

from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection
from typing import List, Union
from voidedlinesreport.model import Line


class OrderRepository(BaseRepository):
    def __init__(self, mbcontext, pos_list):
        # type: (MBEasyContext, List[int]) -> None
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list

    def get_voided_lines(
            self,
            report_pos,         # type: str
            initial_date,       # type: datetime
            end_date,           # type: datetime
            operator_id,        # type: Union[int, None]
            current_session_id  # type: Union[str, None]
    ):
        # type: (...) -> List[Line]

        def inner_func(
                conn    # type: Connection
        ):
            # type: (...) -> List[Line]

            query_filter = """o.SessionId = '{0}'""".format(current_session_id)
            if initial_date is not None:
                query_filter = """
                    (
                        strftime('%Y-%m-%d', oi.LastUpdate, 'localtime') >= '{}' AND 
                        strftime('%Y-%m-%d', oi.LastUpdate, 'localtime') <= '{}'
                    )
                """.format(initial_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d"))
            query = self._GetVoidedLinesQuery.format(query_filter)

            results = [[int(x.get_entry(0) or 0),
                        x.get_entry(1),
                        int(x.get_entry(2) or 0),
                        float(x.get_entry(3) or 0),
                        int(x.get_entry(4) or 0),
                        int(x.get_entry(5) or 0),
                        float(x.get_entry(6) or 0),
                        x.get_entry(7)]
                       for x in conn.select(query)]

            all_lines = []
            for columns in results:
                order_id = columns[0]
                session_id = columns[1]
                line_number = columns[2]
                quantity = columns[3]
                part_code = columns[4]
                price_key = columns[5]
                price = columns[6]
                void_datetime = columns[7]

                operator = int(session_id.split("user=")[1].split(",")[0])
                void_datetime = (
                    None
                    if void_datetime is None
                    else datetime.strptime(void_datetime, "%Y-%m-%dT%H:%M:%S.%f")
                )

                if operator_id and operator_id != operator:
                    continue

                line = Line(order_id, operator, None, line_number, quantity, part_code, price_key, void_datetime, price)
                all_lines.append(line)

            return all_lines

        report_pos_list = self.pos_list if report_pos is None else [report_pos]
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    _GetVoidedLinesQuery = """
    SELECT DISTINCT
        a.OrderId, 
        a.SessionId,
        a.LineNumber,
        CASE
            WHEN a.PartCode == b.PartCode 
            THEN coalesce(a.LastOrderedQty, a.DefaultQty) 
            ELSE coalesce(a.LastOrderedQty, a.DefaultQty) * b.LastOrderedQty 
        END Qty, 	
        a.PartCode, 
        a.PriceKey, 
        a.DefaultUnitPrice,
        a.LastUpdate
    FROM (
        SELECT
            oi.OrderId, 
            o.SessionId, 
            oi.PartCode, 
            oi.PriceKey, 
            oi.LastOrderedQty, 
            oi.DefaultQty, 
            oi.LastUpdate, 
            oi.LineNumber,
            op.DefaultUnitPrice
        FROM Orders o
            JOIN OrderItem oi ON oi.OrderId = o.OrderId
            JOIN OrderPrice op ON op.OrderId = o.OrderId
        WHERE (oi.LastOrderedQty > 0 OR oi.DefaultQty > 0)
            AND oi.PriceKey IS NOT NULL
            AND oi.PriceKey = op.PriceKey
            AND o.StateId in (4, 5, 8)
            AND {}
        ) a
        JOIN OrderItem b ON b.OrderId = a.OrderId 
            AND b.LineNumber = a.LineNumber 
            AND b.Level = 0 
            AND b.LastOrderedQty IS NOT NULL"""



from typing import Any, Tuple

from messageprocessorutil import UuidMessageProcessor

from dailygoals_services.interactor.update_total_sold import UpdateTotalSoldOrderInteractor


class UpdateTotalSoldOrderProcessor(UuidMessageProcessor):
    def __init__(self, interactor):
        # type: (UpdateTotalSoldOrderInteractor) -> None

        super(UpdateTotalSoldOrderProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (Any, str) -> (Any, str)

        return data, event_type

    def call_business(self, params):
        # type: (Tuple) -> None

        data, event_type = params
        self.interactor.update(data, event_type)

    def format_response(self, message_bus, message, event, request, response):
        return

    def format_exception(self, message_bus, message, event, request, exception):
        return

    def format_parse_exception(self, message_bus, message, event, data, exception):
        return



import json

from application.interactor import ApplyBenefitInteractor
from application.model import AsyncUuidMessageProcessor
from messagebus import DefaultToken, DataType, Message, MessageBus, Event
from typing import (
    Optional,
    Any,
    List
)


class ApplyBenefitProcessor(AsyncUuidMessageProcessor):

    def __init__(self, interactor):
        # type: (ApplyBenefitInteractor) -> None

        super(ApplyBenefitProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data):
        # type: (str) -> (int, int, str)

        parsed_data = json.loads(data)
        pos_id = int(parsed_data.get("posId"))
        originator_pos_id = int(parsed_data.get("originatorPosId") or pos_id)
        voucher_id = str(parsed_data.get("voucherId"))
        user_auth_level = parsed_data.get("user_auth_level") or -1

        return pos_id, originator_pos_id, voucher_id, user_auth_level

    def call_business(
            self,
            params      # type: List[int, int, str, bool]
    ):
        # type: (...) -> None

        pos_id, originator_pos_id, voucher_id, user_auth_level = params
        self.interactor.apply_benefit(
                current_pos_id=pos_id,
                originator_pos_id=originator_pos_id,
                voucher_id=voucher_id,
                user_auth_level=user_auth_level
        )

    def format_response(self, message_bus, message, event, request, response):
        # type: (MessageBus, Optional[Message], Optional[Event], Any, Any) -> None

        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, response))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        # type: (MessageBus, Optional[Message], Optional[Event], Any, BaseException) -> None

        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        # type: (MessageBus, Optional[Message], Optional[Event], bytes, BaseException) -> None

        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



from sysactions import (
    action,
    get_model,
    get_posot,
    check_current_order,
    show_messagebox,
    show_confirmation,
    StopAction
)
from posot import OrderTakerException
from custom_sysactions import user_authorized_to_execute_action

from actions.discount.clean_discounts import clean_order_payments
from actions.void.utils import get_void_reason_id, void_order


@action
def do_back_from_total(pos_id, void_reason=5, *args):
    model = get_model(pos_id)
    order = model.find("CurrentOrder/Order")

    try:
        clean_order_payments(
            pos_id=pos_id,
            order=order,
            check_manual_discounts=True
        )
        check_current_order(pos_id, model=model)
        get_posot(model).reopenOrder(int(pos_id))

    except OrderTakerException as ex:
        show_messagebox(pos_id, message="$ERROR_CODE_INFO|%d|%s" % (ex.getErrorCode(), ex.getErrorDescr()))


@user_authorized_to_execute_action
def void_order_with_payments(pos_id, void_reason):
    confirm = show_confirmation(pos_id, "$CANNOT_CHANGE_ORDER_WITH_PAYMENTS", timeout=180000)
    if confirm:
        void_reason = get_void_reason_id(pos_id, void_reason)
        if not void_reason:
            raise StopAction()

        void_order(pos_id, void_reason=void_reason)



class SaleChallenge(object):

    def __init__(self, name, start_date, end_date, products, created_by, challenge_id=""):
        # type: (str, str, str, str, int, int) -> None

        self.name = name
        self.start_date = start_date
        self.end_date = end_date
        self.products = products
        self.created_by = created_by
        self.challenge_id = challenge_id



import time
import random
import struct
import select
import socket


def chk(data):
    x = sum(x << 8 if i % 2 else x for i, x in enumerate(data)) & 0xFFFFFFFF
    x = (x >> 16) + (x & 0xFFFF)
    x = (x >> 16) + (x & 0xFFFF)
    return struct.pack('<H', ~x & 0xFFFF)


def ping(addr, timeout=1, number=1, data=b''):
    with socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_ICMP) as conn:
        payload = struct.pack('!HH', random.randrange(1024, 65536), number) + data

        conn.connect((addr, 80))
        conn.sendall(b'\x08\0' + chk(b'\x08\0\0\0' + payload) + payload)
        start = time.time()

        while select.select([conn], [], [], max(0, start + timeout - time.time()))[0]:
            data = conn.recv(65536)
            if len(data) < 20 or len(data) < struct.unpack_from('!xxH', data)[0]:
                continue
            if data[20:] == b'\0\0' + chk(b'\0\0\0\0' + payload) + payload:
                return time.time() - start



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""Module: barcode.isxn

:Provided barcodes: ISBN-13, ISBN-10, ISSN

This module provides some special codes, which are no standalone barcodes.
All codes where transformed to EAN-13 barcodes. In every case, the checksum
is new calculated.

Example::

    >>> from barcode import get_barcode
    >>> ISBN = get_barcode('isbn10')
    >>> isbn = ISBN('0132354187')
    >>> unicode(isbn)
    u'0132354187'
    >>> isbn.get_fullcode()
    u'9780132354189'
    >>> # Test with wrong checksum
    >>> isbn = ISBN('0132354180')
    >>> unicode(isbn)
    u'0132354187'

"""
__docformat__ = 'restructuredtext en'

from barcode.ean import EuropeanArticleNumber13
from barcode.errors import *


class InternationalStandardBookNumber13(EuropeanArticleNumber13):
    """Initializes new ISBN-13 barcode.

    :parameters:
        isbn : String
            The isbn number as string.
        writer : barcode.writer Instance
            The writer to render the barcode (default: SVGWriter).
    """

    name = 'ISBN-13'

    def __init__(self, isbn, writer=None):
        isbn = isbn.replace('-', '')
        self.isbn13 = isbn
        if isbn[:3] not in ('978', '979'):
            raise WrongCountryCodeError('ISBN must start with 978 or 979.')
        EuropeanArticleNumber13.__init__(self, isbn, writer)


class InternationalStandardBookNumber10(InternationalStandardBookNumber13):
    """Initializes new ISBN-10 barcode. This code is rendered as EAN-13 by
    prefixing it with 978.

    :parameters:
        isbn : String
            The isbn number as string.
        writer : barcode.writer Instance
            The writer to render the barcode (default: SVGWriter).
    """

    name = 'ISBN-10'

    digits = 9

    def __init__(self, isbn, writer=None):
        isbn = isbn.replace('-', '')
        isbn = isbn[:self.digits]
        self.isbn10 = isbn
        self.isbn10 = '{0}{1}'.format(isbn, self._calculate_checksum())
        InternationalStandardBookNumber13.__init__(self, '978' + isbn, writer)

    def _calculate_checksum(self):
        tmp = sum([x * int(y) for x, y in enumerate(self.isbn10[:9],
                                                    start=1)]) % 11
        if tmp == 10:
            return 'X'
        else:
            return tmp

    def __unicode__(self):
        return self.isbn10

    __str__ = __unicode__


class InternationalStandardSerialNumber(EuropeanArticleNumber13):
    """Initializes new ISSN barcode. This code is rendered as EAN-13
    by prefixing it with 977 and adding 00 between code and checksum.

    :parameters:
        issn : String
            The issn number as string.
        writer : barcode.writer Instance
            The writer to render the barcode (default: SVGWriter).
    """

    name = 'ISSN'

    digits = 7

    def __init__(self, issn, writer=None):
        issn = issn.replace('-', '')
        issn = issn[:self.digits]
        self.issn = issn
        self.issn = '{0}{1}'.format(issn, self._calculate_checksum())
        EuropeanArticleNumber13.__init__(self, self.make_ean(), writer)

    def _calculate_checksum(self):
        tmp = 11 - sum([x * int(y) for x, y in
                        enumerate(reversed(self.issn[:7]), start=2)]) % 11
        if tmp == 10:
            return 'X'
        else:
            return tmp

    def make_ean(self):
        return '977{0}00{1}'.format(self.issn[:7], self._calculate_checksum())

    def __unicode__(self):
        return self.issn

    __str__ = __unicode__


# Shortcuts
ISBN13 = InternationalStandardBookNumber13
ISBN10 = InternationalStandardBookNumber10
ISSN = InternationalStandardSerialNumber



import json
import requests

from requests import Response
from typing import Optional, List, Tuple

from helper import retry

from application.model.configuration import Configurations
from application.model.customexception import NotFound, ForbiddenError, ProcessorError, InUseError, \
    NotAvailableError, NotFoundError, NoResponseError, AlreadyUnlockedError
from application.model.orderinfo import Order, OrderJsonEncoder


class APIRepository(object):

    def __init__(self, configs):
        # type: (Configurations) -> None

        self.logger = configs.logger
        self.configs = configs

    def retrieve_loyalty_customer_info(self, loyalty_customer_id):
        # type: (str) -> Optional[str, Exception]
        try:
            url = self._get_customer_url(loyalty_customer_id)
            headers = self._get_headers()
            errors_list = []

            response = retry(
                logger=self.logger,
                method_name="retrieve_loyalty_customer_info",
                n=1,
                sleep_seconds=1,
                trying_to_do=lambda: requests.request("GET", url=url, headers=headers, timeout=20),
                is_success=lambda resp: resp.status_code == 200,
                success=lambda resp: resp,
                failed=lambda resp: errors_list.append((resp.status_code, resp))
            )
            if response:
                return response.content

            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_lock_custom_exception(response_error.status_code, response_error)
        except Exception as ex:
            self.logger.error("Error on get_benefit: {}".format(ex))
            raise

    def retrieve_benefit(self, pos_id, voucher_id):
        # type: (int, str) -> Optional[str, bool]

        order_info = self._get_order_info()

        benefit_raw = self._get_benefit(
            benefit_id=voucher_id,
            order=order_info,
        )

        if not benefit_raw:
            self.logger.info("No benefit found")
            return None

        benefit = json.loads(benefit_raw)
        benefit = benefit.get("benefit", benefit)

        notify = benefit.get("notify", False)
        benefit_id = benefit.get("id", None)
        if notify and benefit_id:
            self._lock_benefit(benefit_id=benefit_id)

        return benefit_raw

    def _get_order_info(self):
        order = Order(self.configs.store_id)
        return json.loads(json.dumps(order, encoding="utf-8", cls=OrderJsonEncoder))

    def burn_benefit(
        self,
        benefit_id,     # type: str
    ):
        # type: (...) -> None

        url = self._get_burn_url(benefit_id)
        headers = self._get_headers()
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="burn_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request("PUT", url=url, headers=headers, timeout=20),
            is_success=lambda resp: resp.status_code == 204,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )
        if response:
            return response.content
        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_unlock_burn_custom_exception(response_error.status_code, response_error)

    def _get_benefit(
        self,
        benefit_id,     # type: str
        order,          # Dict
    ):
        # type: (...) -> str

        url = self._get_benefit_url(voucher_id=benefit_id)
        headers = self._get_headers()
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="get_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request(
                "POST",
                url=url,
                headers=headers,
                timeout=20,
                json=order
            ),
            is_success=lambda resp: resp.status_code == 200,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response:
            return response.content

        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_lock_custom_exception(response_error.status_code, response_error)


    def _lock_benefit(
        self,
        benefit_id  # type: str
    ):
        # type: (...) -> None

        url = self._get_lock_url(benefit_id=benefit_id)
        headers = self._get_headers()
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="lock_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request("PUT", url=url, headers=headers, timeout=20),
            is_success=lambda resp: resp.status_code == 204,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response:
            return response.content
        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list=errors_list)
            self._raise_lock_custom_exception(
                status_code=response_error.status_code,
                response_error=response_error,
            )

    def unlock_benefit(
        self,
        benefit_id,     # type: str
    ):
        # type: (...) -> None

        url = self._get_unlock_url(benefit_id)
        headers = self._get_headers()
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="unlock_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request("PUT", url=url, headers=headers, timeout=20),
            is_success=lambda resp: resp.status_code == 204,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response:
            return response.content
        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_unlock_burn_custom_exception(response_error.status_code, response_error)

    def _get_customer_url(self, loyalty_customer_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.retrieve_loyalty_customer_info.format(customerId=loyalty_customer_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from get customer info endpoint: {}".format(full_url))

        return full_url

    def _get_benefit_url(self, voucher_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.get_benefit.format(voucherId=voucher_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from get benefit endpoint: {}".format(full_url))

        return full_url

    def _get_burn_url(self, benefit_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.burn_benefit.format(benefitId=benefit_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from burn endpoint: {}".format(full_url))

        return full_url

    def _get_lock_url(
        self,
        benefit_id  # type: str
    ):
        # type: (...) -> str

        end_point_url = self.configs.endpoints.lock_benefit.format(benefitId=benefit_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from lock endpoint: {}".format(full_url))

        return full_url

    def _get_unlock_url(self, benefit_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.unlock_benefit.format(benefitId=benefit_id)
        full_url = self.configs.base_url + end_point_url

        self.logger.info("Requesting from unlock endpoint: {}".format(full_url))

        return full_url

    def _get_headers(self):
        # type: () -> dict

        return {"X-Api-Key": self.configs.api_key, "Content-type": "application/json"}

    def _raise_lock_custom_exception(self, status_code, response_error):
        # type: (int, Response) -> None

        if status_code == 204:
            raise NoResponseError(self.logger, response_error)
        if status_code == 403:
            raise ForbiddenError(self.logger, response_error)
        if status_code == 404:
            raise NotFoundError(self.logger, response_error)
        if status_code == 406:
            raise NotAvailableError(self.logger, response_error)
        if status_code == 409:
            raise InUseError(self.logger, response_error)
        if status_code >= 500:
            raise ProcessorError(self.logger, response_error)

    def _raise_unlock_burn_custom_exception(self, status_code, response_error):
        # type: (int, Response) -> None

        if status_code == 204:
            raise NoResponseError(self.logger, response_error)
        if status_code == 400:
            raise InUseError(self.logger, response_error)
        if status_code == 403:
            raise ForbiddenError(self.logger, response_error)
        if status_code == 404:
            raise NotFoundError(self.logger, response_error)
        if status_code == 406:
            raise NotAvailableError(self.logger, response_error)
        if status_code == 409:
            raise AlreadyUnlockedError(self.logger, response_error)
        if status_code >= 500:
            raise ProcessorError(self.logger, response_error)

    @staticmethod
    def _get_response_error(errors_list):
        # type: (List[Tuple]) -> Response

        last_error = errors_list[-1]
        status_code, response = last_error
        return response



from typing import List, Dict, Any
from pos_model import Order

from _SatXmlPartBuilder import SatXmlPartBuilder


class CfeBuilder(SatXmlPartBuilder):
    _cfe_template = u"""<?xml version='1.0' encoding='UTF-8'?><CFe>{0}</CFe>"""

    def __init__(self, part_builders):
        # type: (List[SatXmlPartBuilder]) -> None
        self.part_builders = part_builders

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        inner_xml = u""
        for part_builder in self.part_builders:
            inner_xml += part_builder.build_xml(order, context)

        return self._cfe_template.format(inner_xml)



# -*- coding: utf-8 -*-
import json
import logging
from xml.etree.ElementTree import Element

import persistence
import pyscripts
import re
from datetime import (
    datetime,
    timedelta,
)
from xml.etree import cElementTree as eTree

from bustoken import (
    TK_SLCTRL_OMGR_UPDATEORDERTENDERDETAIL,
    TK_ELECTRONIC_PAYMENT_GET_DOCUMENT,
    TK_GET_USED_SERVICE,
)
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)
from posot import OrderTaker
from sysactions import (
    show_any_dialog,
    show_keyboard,
    show_messagebox,
    get_posot,
    get_model,
    get_current_order,
    translate_message,
    get_cfg,
    get_tender_types,
    StopAction,
)

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

mbcontext = pyscripts.mbcontext
logger = logging.getLogger("TableActions")

cached_tender_types = None
card_brands_cache = None

NOT_ELECTRONIC_TYPE_ID = (0, 9)


def get_void_reason(pos_id, selected_index=None):
    void_reason_ids = range(1, 13)

    void_reason_options = get_void_reason_options(pos_id)
    void_reason_to_show = void_reason_options[0:4]
    if not selected_index:
        selected_index = show_any_dialog(pos_id, "filteredListBox", "", "$SELECT_REASON", "|".join(void_reason_to_show),
                                         "$OK|$CANCEL", "", 600000, "NOFILTER", "", "", "", None, False)

    if selected_index is None:
        return None

    selected_index = int(selected_index)
    void_reason_id = void_reason_ids[int(selected_index)]
    void_reason_descr = void_reason_options[int(selected_index)]
    void_reason_dict = {'VOID_REASON_ID': void_reason_id, 'VOID_REASON_DESCR': void_reason_descr}
    return void_reason_dict


def update_custom_properties(customer_doc=None, customer_name=None, pre_sale=None):
    dict_sale = {}

    if customer_doc is not None:
        dict_sale.update({"CUSTOMER_DOC": customer_doc})
    if customer_name is not None:
        dict_sale.update({"CUSTOMER_NAME": customer_name})
    if pre_sale not in (None, ""):
        dict_sale.update({"PRE_VENDA": pre_sale})

    return dict_sale


def get_customer_doc_after_paid(
    pos_id,                     # type: str
    pos_ot,                     # type: OrderTaker
    ask_document_pinpad=False   # type: Optional[bool]
):
    # type: (...) -> tuple

    customer_info_config = get_customer_info_config(pos_id)

    document = get_custom_property_value(pos_id, "CUSTOMER_DOC")
    if document:
        return document, False

    if "afterPaid" not in customer_info_config['document']:
        return "", True

    valid = False
    customer_doc = None
    while not valid:
        if ask_document_pinpad:
            customer_doc = input_document_pinpad_dialog(pos_id)

        if customer_doc is None:
            customer_doc = show_keyboard(pos_id, "$ONLY_NUMBER", "$ENTER_CPF_CNPJ", mask="CPF_CNPJ", numpad=True)

        if customer_doc is None:
            return None, False

        if customer_doc == "":
            ret = show_messagebox(pos_id, "$CONFIRM_EMPTY_CPF_CNPJ", buttons="$YES|$NO")
            if ret == 1:
                continue
            break
        customer_doc = str(customer_doc)
        valid = validate_document(customer_doc)
        if not valid:
            show_messagebox(pos_id, "$INVALID_CPF_CNPJ")

    fill_additional_info(pos_ot, "CPF", customer_doc)
    return customer_doc, True


def electronic_payment_get_document(
    pos_id      # type: str
):
    # type: (...) -> Optional[str]

    component_name = get_device_service(
        pos_id=pos_id,
        device_type="eft",
    )
    if not component_name:
        return None

    try:
        result = mbcontext.MB_EasySendMessage(
            dest_name=component_name,
            token=TK_ELECTRONIC_PAYMENT_GET_DOCUMENT,
            data=pos_id
        )
        if result.token == TK_SYS_ACK:
            return result.data

        return None

    except Exception as _:
        logger.exception("Error ask documents from pinpad")
        return None


def input_document_pinpad_dialog(
    pos_id  # type: str
):
    # type: (...) -> Optional[str]

    option = show_messagebox(
        pos_id,
        message="$ASK_CPF_CNPJ",
        buttons="$CANCEL_WITHOUT_ICON|$NO_WITHOUT_ICON|$MANUAL|$PINPAD"
    )

    if option == 0:
        raise StopAction()
    elif option == 1:
        customer_doc = ""
    elif option == 2:
        customer_doc = None
    else:
        customer_doc = electronic_payment_get_document(pos_id)

    return customer_doc


def validate_document(customer_doc):
    valid = False
    if len(customer_doc) <= 11:
        valid = validar_cpf(customer_doc)
    elif len(customer_doc) > 11:
        valid = validar_cnpj(customer_doc)
    return valid


def fill_additional_info(pos_ot, properties, value):
    if pos_ot.additionalInfo:
        pos_ot.additionalInfo += "|"
    else:
        pos_ot.additionalInfo = ""
    pos_ot.additionalInfo += "{0}={1}".format(properties, value)


def validar_cpf(cpf):
    cpf = ''.join(re.findall('\d', str(cpf)))

    if (not cpf) or (len(cpf) < 11):
        return False

    inteiros = map(int, cpf)
    novo = inteiros[:9]

    if all(num == inteiros[0] for num in inteiros):
        return False

    while len(novo) < 11:
        r = sum([(len(novo) + 1 - i) * v for i, v in enumerate(novo)]) % 11

        if r > 1:
            f = 11 - r
        else:
            f = 0
        novo.append(f)

    if novo == inteiros:
        return cpf

    return False


def validar_cnpj(cnpj):
    cnpj = ''.join(re.findall('\d', str(cnpj)))

    if (not cnpj) or (len(cnpj) < 14):
        return False

    inteiros = map(int, cnpj)
    novo = inteiros[:12]

    if all(num == inteiros[0] for num in inteiros):
        return False

    prod = [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
    while len(novo) < 14:
        r = sum([x * y for (x, y) in zip(novo, prod)]) % 11
        if r > 1:
            f = 11 - r
        else:
            f = 0
        novo.append(f)
        prod.insert(0, 6)

    if novo == inteiros:
        return cnpj

    return False


def json_serialize(dictionary):
    # type: (Dict) -> str
    try:
        return json.dumps(dictionary, default=lambda o: getattr(o, '__dict__', str(o)))
    except ValueError as _:
        return dictionary


def compare_objects(obj1, obj2):
    return object_to_json(obj1) == object_to_json(obj2)


def object_to_json(obj):
    return json.dumps(obj, sort_keys=True, default=lambda o: getattr(o, '__dict__', str(o)))


def get_seconds_to_renew_cache(seconds_to_renew_cache):
    seconds_to_renew = 60 * 5
    try:
        seconds_to_renew = int(seconds_to_renew_cache)
    except TypeError as _:
        pass
    return seconds_to_renew


def reached_seconds_to_renew_cache(last_cache_time, seconds_to_renew_cache):
    return not last_cache_time or datetime.now() > last_cache_time + timedelta(seconds=seconds_to_renew_cache)


def get_custom_property_value(
    pos_id,              # type: str
    custom_property,     # type: str
    order_id=None,       # type: Optional[str]
    order_picture=None   # type: eTree.XML
):
    # type: (...) -> Optional[str]

    model = get_model(pos_id)
    pos_ot = get_posot(model)

    if not order_picture:
        if not order_id:
            order_picture = get_current_order(model)
        else:
            order_picture = eTree.XML(pos_ot.orderPicture(pos_id, order_id))

    current_property = order_picture.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format(custom_property))
    return current_property if current_property is None else current_property.get('value')

def mask_phone(
        phone_number  # type: str
):
    # type: (...) -> str

    phone_number_size = len(phone_number)

    if phone_number_size in (8, 9):
        return "{}-{}".format(
            phone_number[:-4],
            phone_number[-4:]
        )
    if phone_number_size in (10, 11):
        return "({}) {}-{}".format(
            phone_number[0:2],
            phone_number[2:-4],
            phone_number[-4:]
        )

    return phone_number


def is_valid_string(
    string,             # type: str
    match_string=None   # type: Optional[str]
):
    # type: (...) -> bool

    if not match_string:
        match_string = '^[_a-zA-Z0-9 ,.-]+$'
    return bool(re.match(match_string, string))


def get_card_brands():
    # type: () -> Dict

    global card_brands_cache
    if card_brands_cache is not None:
        return card_brands_cache

    card_brands = {}
    conn = None
    try:
        conn = persistence.Driver().open(
            mbcontext,
            service_name="FiscalPersistence"
        )
        cursor = conn.pselect("fiscal_getBandeiras")
        for row in cursor:
            card_brands.update({int(row.get_entry("Bandeira")): row.get_entry("Descricao")})
    except Exception as _:
        logger.exception("Error getting cards brands")
    finally:
        if conn:
            conn.close()

    card_brands_cache = card_brands
    return card_brands_cache


def format_currency(
    model,      # type: eTree.XML
    value       # type: float
):
    # type: (...) -> str

    currency_symbol = translate_message(
        model,
        "L10N_CURRENCY_SYMBOL"
    )

    value_str = "{0} {1:.2f}".format(
        currency_symbol,
        value
    )
    return value_str


def _count_option_listbox(
    option_text,    # type: str
    value,          # type: float
    options_dict    # type: List[Dict]
):
    # type: (...) -> int

    count = 0
    for option in options_dict:
        if option_text == option["text"] and value == option["value"]:
            count += 1

    return count


def options_dict_to_str_list(
    model,      # type: eTree.XML
    options     # type: List[Dict]
):
    # type: (...) -> List[str]

    listbox = []
    for option in options:
        formatted_value = format_currency(
            model,
            option["value"]
        )
        count = _count_option_listbox(
            option["text"],
            option["value"],
            options
        )
        if count > 1:
            option_text = "{0}: {1} ({2})".format(
                option["text"],
                formatted_value,
                option["count"]
            )
        else:
            option_text = "{0}: {1}".format(
                option["text"],
                formatted_value
            )
        listbox.append(option_text)

    return listbox


def check_has_tef(
    tenders,    # type: eTree.XML
):
    # type: (...) -> bool

    for tender in tenders:
        if is_electronic_payment(tender):
            return True

    return False


def get_tenders_options_dict(
    pos_id,                         # type: str
    model,                          # type: eTree.XML
    tenders,                        # type: eTree.XML
    total,                          # type: float
    card_brands,                    # type: Dict[int, str]
    group_digital_payments=False    # type: bool
):
    # type: (...) -> Tuple[List[Dict], List]

    all_tenders_option = {
        "text": translate_message(
            model,
            "ALL_TENDERS"
        ),
        "value": total,
        "count": 1
    }

    disabled_items_index = []
    tef_not_available = get_device_service(
        pos_id=pos_id,
        device_type="eft",
    ) is None

    options_dict = [all_tenders_option]
    if group_digital_payments and check_has_tef(tenders):
        digital_payments_option = {
            "text": translate_message(
                model,
                "ELECTRONIC_PAYMENTS"
            ),
            "value": 0,
            "count": 1,
            "tenders": []
        }
        options_dict.append(digital_payments_option)

    option_text = ""
    for tender_index, tender in enumerate(tenders):
        value = float(tender.get("tenderAmount"))
        if value == 0:
            continue

        is_tef = is_electronic_payment(tender)
        if is_tef and tef_not_available:
            disabled_items_index.append(str(tender_index + 1))

        if is_tef:
            if not group_digital_payments:
                option_text = _get_option_text(
                    tender,
                    card_brands
                )
            else:
                options_dict[1]["value"] += float(value)
                options_dict[1]["tenders"].append(tender.get("tenderId"))
        else:
            option_text = tender.get("tenderDescr").encode("utf-8")

        if not is_tef or (is_tef and not group_digital_payments):
            count = _count_option_listbox(
                option_text,
                value,
                options_dict
            )
            new_option = {
                "text": option_text,
                "tender_index": tender_index,
                "value": value,
                "count": count + 1
            }
            options_dict.append(new_option)

    if len(disabled_items_index) > 0:
        disabled_items_index.insert(0, '0')

    return options_dict, disabled_items_index


def _get_option_text(
    tender,         # type: eTree.Element
    card_brands,    # type: Dict[int, str]
):
    detail = json.loads(tender.get("tenderDetail"))
    tender_description = tender.get("tenderDescr").encode("utf-8")
    brand_code = int(detail["Bandeira"])
    brand = card_brands[int(detail["Bandeira"])] if brand_code in card_brands else ""
    card_number = str(detail["CardNumber"][:4] if 'None' not in detail["CardNumber"][:4] else None)

    if card_number in ['', None]:
        return "{0} / {1}".format(
            tender_description,
            brand
        )

    return "{0} / {1} / {2}".format(
            tender_description,
            card_number,
            brand
    )


def get_order_tender(
    order_picture,  # type: eTree.XML
    tender_id       # type: str
):
    # type: (...) -> Optional[eTree.XML]

    tenders = order_picture.findall(".//TenderHistory/Tender")
    for tender in tenders:
        if tender.get("tenderId") == tender_id:
            return tender

    return None


def save_order_cancel_receipt(
    pos_id,         # type: str
    tender_id       # type: str
):
    # type: (...) -> bool

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order_picture = get_current_order(model)
    tender = get_order_tender(
        order_picture,
        tender_id
    )
    tender_details = json.loads(tender.get("tenderDetail"))
    receipt_cancel = {
        "posId": pos_id,
        "userId": model.find("Operator").get("id"),
        "canceledAt": str(datetime.now())
    }

    if "ReceiptCancel" in tender_details:
        tender_details["ReceiptCancel"].update(receipt_cancel)
    else:
        tender_details["ReceiptCancel"] = receipt_cancel

    try:
        params = (
            '1' if pos_ot.blkopnotify else '0',
            pos_id,
            order_picture.get("orderId"),
            json.dumps([{
                "tenderid": int(tender.get("tenderId")),
                "tenderdetail": json.dumps(tender_details)
            }])
        )
        req_buf = "\x00".join((str(p) for p in params))
        result = mbcontext.MB_EasySendMessage(
            "ORDERMGR%s" % pos_id,
            TK_SLCTRL_OMGR_UPDATEORDERTENDERDETAIL,
            FM_PARAM,
            req_buf,
            600000000
        )
        if result.token == TK_SYS_ACK:
            return True

    except Exception as _:
        logger.exception("Fail salve receipt cancel: Tender: {}".format(tender.get("tenderId")))

    return False


def is_electronic_payment(
    tender   # type: Element
):
    # type: (...) -> bool

    return _get_electronic_type_id(tender) not in NOT_ELECTRONIC_TYPE_ID


def _get_electronic_type_id(
    tender   # type: Element
):
    # type: (...) -> int

    if type(tender) is dict:
        return int(tender.get("electronicTypeId"))

    return int(tender.get("electronicType") or tender.get("eletronicType"))


def get_customer_info_config(pos_id):
    document = get_cfg(pos_id).find_values("CustomerInfoRequest/Document") or []
    name = get_cfg(pos_id).find_values("CustomerInfoRequest/Name") or []

    return {'document': document, 'name': name}


def get_void_reason_options(
    pos_id  # type: str
):
    # type: (...) -> List[str]

    void_reason_options = ["CHANGED_YOUR_MIND",
                           "DUPLICATE",
                           "WRONG_SALE",
                           "CANCELLATION",
                           "COUPON_CANCELED",
                           "PAYMENT_CANCELED",
                           "SAVED_ORDER_CANCELED",
                           "ERROR_CREATING_SALE",
                           "ERROR_PROCESSING_PAYMENT_SELECTION",
                           "CANCELLATION_VIA_SYSTEM",
                           "SYNC_ERROR",
                           "MANAGER_AUTHORIZATION_CANCELLATION"]

    return _translate_array(pos_id, void_reason_options)


def _translate_array(pos_id, messages):
    # type: (str, List[str]) -> List[str]

    model = get_model(pos_id)
    result = []
    for index, message in enumerate(messages, start=1):
        result.append("{} - {}".format(index, translate_message(model, message)))

    return result


def get_tender_info(
    tender_id=None,     # type: Optional[int]
):
    # type: (...) -> Optional[Any[Union[Dict, object]]]

    global cached_tender_types

    if cached_tender_types is None:
        cached_tender_types = get_tender_types()

    if tender_id is None:
        return cached_tender_types

    return cached_tender_types[tender_id] if tender_id in cached_tender_types else None


def _get_substring_occurrence(
    text,       # type: str
    character,  # type: str
):
    # type: (...) -> str

    first_occurrence = text.find(character)
    return character * (text.rfind(character) - first_occurrence + 1 if first_occurrence > -1 else 0)


def pad_barcode(
    barcode,   # type: str
):
    # type: (...) -> str

    barcode = barcode.replace(' ', '')
    barcode_2 = _get_substring_occurrence(
        text=barcode,
        character="2",
    )
    barcode_x = _get_substring_occurrence(
        text=barcode,
        character="X",
    )
    barcode_y = _get_substring_occurrence(
        text=barcode,
        character="Y",
    )
    barcode_w = _get_substring_occurrence(
        text=barcode,
        character="W",
    )

    return '{}{}{}{}{}'.format(
        barcode_2,
        barcode_x,
        '0' * (13 - (len(barcode_2) + len(barcode_x) + len(barcode_y) + len(barcode_w))),
        barcode_y,
        barcode_w
    )


def get_sector_name(
    order,  # type: eTree
):
    # type: (...) -> Optional[str]

    sector = order.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("SECTOR"))
    if sector is not None:
        return sector.get("value", None)

    table_id = order.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("TABLE_ID"))
    if table_id is not None and "TAB-" in table_id.get("value", None):
        return "TAB"

    return None


def get_device_service(
    pos_id,         # type: str
    device_type,    # type: str
    sector=None,    # type: str
):
    # type: (...) -> str

    request = json.dumps(
        {
            "posId": pos_id,
            "deviceType": device_type,
            "sector": sector,
        }
    )

    msg = mbcontext.MB_EasySendMessage(
        dest_name="DeviceManager",
        token=TK_GET_USED_SERVICE,
        format=FM_PARAM,
        data=request,
    )

    if msg.token == TK_SYS_ACK:
        return msg.data

    raise Exception(msg.data)



from externals.util import encode_utf_8

from ._Address import Address


class Customer(object):
    def __init__(self, phone, document, name, address):
        # type: (str, str, str, Address) -> None

        self.phone = phone
        self.document = document
        self.name = encode_utf_8(name)
        self.address = address



# -*- coding: utf-8 -*-

"""
requests.utils
~~~~~~~~~~~~~~

This module provides utility functions that are used within Requests
that are also useful for external consumption.
"""

import codecs
import contextlib
import io
import os
import re
import socket
import struct
import sys
import tempfile
import warnings
import zipfile
from collections import OrderedDict

from .__version__ import __version__
from . import certs
# to_native_string is unused here, but imported here for backwards compatibility
from ._internal_utils import to_native_string
from .compat import parse_http_list as _parse_list_header
from .compat import (
    quote, urlparse, bytes, str, unquote, getproxies,
    proxy_bypass, urlunparse, basestring, integer_types, is_py3,
    proxy_bypass_environment, getproxies_environment, Mapping)
from .cookies import cookiejar_from_dict
from .structures import CaseInsensitiveDict
from .exceptions import (
    InvalidURL, InvalidHeader, FileModeWarning, UnrewindableBodyError)

NETRC_FILES = ('.netrc', '_netrc')

DEFAULT_CA_BUNDLE_PATH = certs.where()

DEFAULT_PORTS = {'http': 80, 'https': 443}


if sys.platform == 'win32':
    # provide a proxy_bypass version on Windows without DNS lookups

    def proxy_bypass_registry(host):
        try:
            if is_py3:
                import winreg
            else:
                import _winreg as winreg
        except ImportError:
            return False

        try:
            internetSettings = winreg.OpenKey(winreg.HKEY_CURRENT_USER,
                r'Software\Microsoft\Windows\CurrentVersion\Internet Settings')
            # ProxyEnable could be REG_SZ or REG_DWORD, normalizing it
            proxyEnable = int(winreg.QueryValueEx(internetSettings,
                                              'ProxyEnable')[0])
            # ProxyOverride is almost always a string
            proxyOverride = winreg.QueryValueEx(internetSettings,
                                                'ProxyOverride')[0]
        except OSError:
            return False
        if not proxyEnable or not proxyOverride:
            return False

        # make a check value list from the registry entry: replace the
        # '<local>' string by the localhost entry and the corresponding
        # canonical entry.
        proxyOverride = proxyOverride.split(';')
        # now check if we match one of the registry values.
        for test in proxyOverride:
            if test == '<local>':
                if '.' not in host:
                    return True
            test = test.replace(".", r"\.")     # mask dots
            test = test.replace("*", r".*")     # change glob sequence
            test = test.replace("?", r".")      # change glob char
            if re.match(test, host, re.I):
                return True
        return False

    def proxy_bypass(host):  # noqa
        """Return True, if the host should be bypassed.

        Checks proxy settings gathered from the environment, if specified,
        or the registry.
        """
        if getproxies_environment():
            return proxy_bypass_environment(host)
        else:
            return proxy_bypass_registry(host)


def dict_to_sequence(d):
    """Returns an internal sequence dictionary update."""

    if hasattr(d, 'items'):
        d = d.items()

    return d


def super_len(o):
    total_length = None
    current_position = 0

    if hasattr(o, '__len__'):
        total_length = len(o)

    elif hasattr(o, 'len'):
        total_length = o.len

    elif hasattr(o, 'fileno'):
        try:
            fileno = o.fileno()
        except io.UnsupportedOperation:
            pass
        else:
            total_length = os.fstat(fileno).st_size

            # Having used fstat to determine the file length, we need to
            # confirm that this file was opened up in binary mode.
            if 'b' not in o.mode:
                warnings.warn((
                    "Requests has determined the content-length for this "
                    "request using the binary size of the file: however, the "
                    "file has been opened in text mode (i.e. without the 'b' "
                    "flag in the mode). This may lead to an incorrect "
                    "content-length. In Requests 3.0, support will be removed "
                    "for files in text mode."),
                    FileModeWarning
                )

    if hasattr(o, 'tell'):
        try:
            current_position = o.tell()
        except (OSError, IOError):
            # This can happen in some weird situations, such as when the file
            # is actually a special file descriptor like stdin. In this
            # instance, we don't know what the length is, so set it to zero and
            # let requests chunk it instead.
            if total_length is not None:
                current_position = total_length
        else:
            if hasattr(o, 'seek') and total_length is None:
                # StringIO and BytesIO have seek but no useable fileno
                try:
                    # seek to end of file
                    o.seek(0, 2)
                    total_length = o.tell()

                    # seek back to current position to support
                    # partially read file-like objects
                    o.seek(current_position or 0)
                except (OSError, IOError):
                    total_length = 0

    if total_length is None:
        total_length = 0

    return max(0, total_length - current_position)


def get_netrc_auth(url, raise_errors=False):
    """Returns the Requests tuple auth for a given url from netrc."""

    netrc_file = os.environ.get('NETRC')
    if netrc_file is not None:
        netrc_locations = (netrc_file,)
    else:
        netrc_locations = ('~/{}'.format(f) for f in NETRC_FILES)

    try:
        from netrc import netrc, NetrcParseError

        netrc_path = None

        for f in netrc_locations:
            try:
                loc = os.path.expanduser(f)
            except KeyError:
                # os.path.expanduser can fail when $HOME is undefined and
                # getpwuid fails. See https://bugs.python.org/issue20164 &
                # https://github.com/psf/requests/issues/1846
                return

            if os.path.exists(loc):
                netrc_path = loc
                break

        # Abort early if there isn't one.
        if netrc_path is None:
            return

        ri = urlparse(url)

        # Strip port numbers from netloc. This weird `if...encode`` dance is
        # used for Python 3.2, which doesn't support unicode literals.
        splitstr = b':'
        if isinstance(url, str):
            splitstr = splitstr.decode('ascii')
        host = ri.netloc.split(splitstr)[0]

        try:
            _netrc = netrc(netrc_path).authenticators(host)
            if _netrc:
                # Return with login / password
                login_i = (0 if _netrc[0] else 1)
                return (_netrc[login_i], _netrc[2])
        except (NetrcParseError, IOError):
            # If there was a parsing error or a permissions issue reading the file,
            # we'll just skip netrc auth unless explicitly asked to raise errors.
            if raise_errors:
                raise

    # App Engine hackiness.
    except (ImportError, AttributeError):
        pass


def guess_filename(obj):
    """Tries to guess the filename of the given object."""
    name = getattr(obj, 'name', None)
    if (name and isinstance(name, basestring) and name[0] != '<' and
            name[-1] != '>'):
        return os.path.basename(name)


def extract_zipped_paths(path):
    """Replace nonexistent paths that look like they refer to a member of a zip
    archive with the location of an extracted copy of the target, or else
    just return the provided path unchanged.
    """
    if os.path.exists(path):
        # this is already a valid path, no need to do anything further
        return path

    # find the first valid part of the provided path and treat that as a zip archive
    # assume the rest of the path is the name of a member in the archive
    archive, member = os.path.split(path)
    while archive and not os.path.exists(archive):
        archive, prefix = os.path.split(archive)
        member = '/'.join([prefix, member])

    if not zipfile.is_zipfile(archive):
        return path

    zip_file = zipfile.ZipFile(archive)
    if member not in zip_file.namelist():
        return path

    # we have a valid zip archive and a valid member of that archive
    tmp = tempfile.gettempdir()
    extracted_path = os.path.join(tmp, *member.split('/'))
    if not os.path.exists(extracted_path):
        extracted_path = zip_file.extract(member, path=tmp)

    return extracted_path


def from_key_val_list(value):
    """Take an object and test to see if it can be represented as a
    dictionary. Unless it can not be represented as such, return an
    OrderedDict, e.g.,

    ::

        >>> from_key_val_list([('key', 'val')])
        OrderedDict([('key', 'val')])
        >>> from_key_val_list('string')
        Traceback (most recent call last):
        ...
        ValueError: cannot encode objects that are not 2-tuples
        >>> from_key_val_list({'key': 'val'})
        OrderedDict([('key', 'val')])

    :rtype: OrderedDict
    """
    if value is None:
        return None

    if isinstance(value, (str, bytes, bool, int)):
        raise ValueError('cannot encode objects that are not 2-tuples')

    return OrderedDict(value)


def to_key_val_list(value):
    """Take an object and test to see if it can be represented as a
    dictionary. If it can be, return a list of tuples, e.g.,

    ::

        >>> to_key_val_list([('key', 'val')])
        [('key', 'val')]
        >>> to_key_val_list({'key': 'val'})
        [('key', 'val')]
        >>> to_key_val_list('string')
        Traceback (most recent call last):
        ...
        ValueError: cannot encode objects that are not 2-tuples

    :rtype: list
    """
    if value is None:
        return None

    if isinstance(value, (str, bytes, bool, int)):
        raise ValueError('cannot encode objects that are not 2-tuples')

    if isinstance(value, Mapping):
        value = value.items()

    return list(value)


# From mitsuhiko/werkzeug (used with permission).
def parse_list_header(value):
    """Parse lists as described by RFC 2068 Section 2.

    In particular, parse comma-separated lists where the elements of
    the list may include quoted-strings.  A quoted-string could
    contain a comma.  A non-quoted string could have quotes in the
    middle.  Quotes are removed automatically after parsing.

    It basically works like :func:`parse_set_header` just that items
    may appear multiple times and case sensitivity is preserved.

    The return value is a standard :class:`list`:

    >>> parse_list_header('token, "quoted value"')
    ['token', 'quoted value']

    To create a header from the :class:`list` again, use the
    :func:`dump_header` function.

    :param value: a string with a list header.
    :return: :class:`list`
    :rtype: list
    """
    result = []
    for item in _parse_list_header(value):
        if item[:1] == item[-1:] == '"':
            item = unquote_header_value(item[1:-1])
        result.append(item)
    return result


# From mitsuhiko/werkzeug (used with permission).
def parse_dict_header(value):
    """Parse lists of key, value pairs as described by RFC 2068 Section 2 and
    convert them into a python dict:

    >>> d = parse_dict_header('foo="is a fish", bar="as well"')
    >>> type(d) is dict
    True
    >>> sorted(d.items())
    [('bar', 'as well'), ('foo', 'is a fish')]

    If there is no value for a key it will be `None`:

    >>> parse_dict_header('key_without_value')
    {'key_without_value': None}

    To create a header from the :class:`dict` again, use the
    :func:`dump_header` function.

    :param value: a string with a dict header.
    :return: :class:`dict`
    :rtype: dict
    """
    result = {}
    for item in _parse_list_header(value):
        if '=' not in item:
            result[item] = None
            continue
        name, value = item.split('=', 1)
        if value[:1] == value[-1:] == '"':
            value = unquote_header_value(value[1:-1])
        result[name] = value
    return result


# From mitsuhiko/werkzeug (used with permission).
def unquote_header_value(value, is_filename=False):
    r"""Unquotes a header value.  (Reversal of :func:`quote_header_value`).
    This does not use the real unquoting but what browsers are actually
    using for quoting.

    :param value: the header value to unquote.
    :rtype: str
    """
    if value and value[0] == value[-1] == '"':
        # this is not the real unquoting, but fixing this so that the
        # RFC is met will result in bugs with internet explorer and
        # probably some other browsers as well.  IE for example is
        # uploading files with "C:\foo\bar.txt" as filename
        value = value[1:-1]

        # if this is a filename and the starting characters look like
        # a UNC path, then just return the value without quotes.  Using the
        # replace sequence below on a UNC path has the effect of turning
        # the leading double slash into a single slash and then
        # _fix_ie_filename() doesn't work correctly.  See #458.
        if not is_filename or value[:2] != '\\\\':
            return value.replace('\\\\', '\\').replace('\\"', '"')
    return value


def dict_from_cookiejar(cj):
    """Returns a key/value dictionary from a CookieJar.

    :param cj: CookieJar object to extract cookies from.
    :rtype: dict
    """

    cookie_dict = {}

    for cookie in cj:
        cookie_dict[cookie.name] = cookie.value

    return cookie_dict


def add_dict_to_cookiejar(cj, cookie_dict):
    """Returns a CookieJar from a key/value dictionary.

    :param cj: CookieJar to insert cookies into.
    :param cookie_dict: Dict of key/values to insert into CookieJar.
    :rtype: CookieJar
    """

    return cookiejar_from_dict(cookie_dict, cj)


def get_encodings_from_content(content):
    """Returns encodings from given content string.

    :param content: bytestring to extract encodings from.
    """
    warnings.warn((
        'In requests 3.0, get_encodings_from_content will be removed. For '
        'more information, please see the discussion on issue #2266. (This'
        ' warning should only appear once.)'),
        DeprecationWarning)

    charset_re = re.compile(r'<meta.*?charset=["\']*(.+?)["\'>]', flags=re.I)
    pragma_re = re.compile(r'<meta.*?content=["\']*;?charset=(.+?)["\'>]', flags=re.I)
    xml_re = re.compile(r'^<\?xml.*?encoding=["\']*(.+?)["\'>]')

    return (charset_re.findall(content) +
            pragma_re.findall(content) +
            xml_re.findall(content))


def _parse_content_type_header(header):
    """Returns content type and parameters from given header

    :param header: string
    :return: tuple containing content type and dictionary of
         parameters
    """

    tokens = header.split(';')
    content_type, params = tokens[0].strip(), tokens[1:]
    params_dict = {}
    items_to_strip = "\"' "

    for param in params:
        param = param.strip()
        if param:
            key, value = param, True
            index_of_equals = param.find("=")
            if index_of_equals != -1:
                key = param[:index_of_equals].strip(items_to_strip)
                value = param[index_of_equals + 1:].strip(items_to_strip)
            params_dict[key.lower()] = value
    return content_type, params_dict


def get_encoding_from_headers(headers):
    """Returns encodings from given HTTP Header Dict.

    :param headers: dictionary to extract encoding from.
    :rtype: str
    """

    content_type = headers.get('content-type')

    if not content_type:
        return None

    content_type, params = _parse_content_type_header(content_type)

    if 'charset' in params:
        return params['charset'].strip("'\"")

    if 'text' in content_type:
        return 'ISO-8859-1'

    if 'application/json' in content_type:
        # Assume UTF-8 based on RFC 4627: https://www.ietf.org/rfc/rfc4627.txt since the charset was unset
        return 'utf-8'


def stream_decode_response_unicode(iterator, r):
    """Stream decodes a iterator."""

    if r.encoding is None:
        for item in iterator:
            yield item
        return

    decoder = codecs.getincrementaldecoder(r.encoding)(errors='replace')
    for chunk in iterator:
        rv = decoder.decode(chunk)
        if rv:
            yield rv
    rv = decoder.decode(b'', final=True)
    if rv:
        yield rv


def iter_slices(string, slice_length):
    """Iterate over slices of a string."""
    pos = 0
    if slice_length is None or slice_length <= 0:
        slice_length = len(string)
    while pos < len(string):
        yield string[pos:pos + slice_length]
        pos += slice_length


def get_unicode_from_response(r):
    """Returns the requested content back in unicode.

    :param r: Response object to get unicode content from.

    Tried:

    1. charset from content-type
    2. fall back and replace all unicode characters

    :rtype: str
    """
    warnings.warn((
        'In requests 3.0, get_unicode_from_response will be removed. For '
        'more information, please see the discussion on issue #2266. (This'
        ' warning should only appear once.)'),
        DeprecationWarning)

    tried_encodings = []

    # Try charset from content-type
    encoding = get_encoding_from_headers(r.headers)

    if encoding:
        try:
            return str(r.content, encoding)
        except UnicodeError:
            tried_encodings.append(encoding)

    # Fall back:
    try:
        return str(r.content, encoding, errors='replace')
    except TypeError:
        return r.content


# The unreserved URI characters (RFC 3986)
UNRESERVED_SET = frozenset(
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz" + "0123456789-._~")


def unquote_unreserved(uri):
    """Un-escape any percent-escape sequences in a URI that are unreserved
    characters. This leaves all reserved, illegal and non-ASCII bytes encoded.

    :rtype: str
    """
    parts = uri.split('%')
    for i in range(1, len(parts)):
        h = parts[i][0:2]
        if len(h) == 2 and h.isalnum():
            try:
                c = chr(int(h, 16))
            except ValueError:
                raise InvalidURL("Invalid percent-escape sequence: '%s'" % h)

            if c in UNRESERVED_SET:
                parts[i] = c + parts[i][2:]
            else:
                parts[i] = '%' + parts[i]
        else:
            parts[i] = '%' + parts[i]
    return ''.join(parts)


def requote_uri(uri):
    """Re-quote the given URI.

    This function passes the given URI through an unquote/quote cycle to
    ensure that it is fully and consistently quoted.

    :rtype: str
    """
    safe_with_percent = "!#$%&'()*+,/:;=?@[]~"
    safe_without_percent = "!#$&'()*+,/:;=?@[]~"
    try:
        # Unquote only the unreserved characters
        # Then quote only illegal characters (do not quote reserved,
        # unreserved, or '%')
        return quote(unquote_unreserved(uri), safe=safe_with_percent)
    except InvalidURL:
        # We couldn't unquote the given URI, so let's try quoting it, but
        # there may be unquoted '%'s in the URI. We need to make sure they're
        # properly quoted so they do not cause issues elsewhere.
        return quote(uri, safe=safe_without_percent)


def address_in_network(ip, net):
    """This function allows you to check if an IP belongs to a network subnet

    Example: returns True if ip = 192.168.1.1 and net = 192.168.1.0/24
             returns False if ip = 192.168.1.1 and net = 192.168.100.0/24

    :rtype: bool
    """
    ipaddr = struct.unpack('=L', socket.inet_aton(ip))[0]
    netaddr, bits = net.split('/')
    netmask = struct.unpack('=L', socket.inet_aton(dotted_netmask(int(bits))))[0]
    network = struct.unpack('=L', socket.inet_aton(netaddr))[0] & netmask
    return (ipaddr & netmask) == (network & netmask)


def dotted_netmask(mask):
    """Converts mask from /xx format to xxx.xxx.xxx.xxx

    Example: if mask is 24 function returns 255.255.255.0

    :rtype: str
    """
    bits = 0xffffffff ^ (1 << 32 - mask) - 1
    return socket.inet_ntoa(struct.pack('>I', bits))


def is_ipv4_address(string_ip):
    """
    :rtype: bool
    """
    try:
        socket.inet_aton(string_ip)
    except socket.error:
        return False
    return True


def is_valid_cidr(string_network):
    """
    Very simple check of the cidr format in no_proxy variable.

    :rtype: bool
    """
    if string_network.count('/') == 1:
        try:
            mask = int(string_network.split('/')[1])
        except ValueError:
            return False

        if mask < 1 or mask > 32:
            return False

        try:
            socket.inet_aton(string_network.split('/')[0])
        except socket.error:
            return False
    else:
        return False
    return True


@contextlib.contextmanager
def set_environ(env_name, value):
    """Set the environment variable 'env_name' to 'value'

    Save previous value, yield, and then restore the previous value stored in
    the environment variable 'env_name'.

    If 'value' is None, do nothing"""
    value_changed = value is not None
    if value_changed:
        old_value = os.environ.get(env_name)
        os.environ[env_name] = value
    try:
        yield
    finally:
        if value_changed:
            if old_value is None:
                del os.environ[env_name]
            else:
                os.environ[env_name] = old_value


def should_bypass_proxies(url, no_proxy):
    """
    Returns whether we should bypass proxies or not.

    :rtype: bool
    """
    # Prioritize lowercase environment variables over uppercase
    # to keep a consistent behaviour with other http projects (curl, wget).
    get_proxy = lambda k: os.environ.get(k) or os.environ.get(k.upper())

    # First check whether no_proxy is defined. If it is, check that the URL
    # we're getting isn't in the no_proxy list.
    no_proxy_arg = no_proxy
    if no_proxy is None:
        no_proxy = get_proxy('no_proxy')
    parsed = urlparse(url)

    if parsed.hostname is None:
        # URLs don't always have hostnames, e.g. file:/// urls.
        return True

    if no_proxy:
        # We need to check whether we match here. We need to see if we match
        # the end of the hostname, both with and without the port.
        no_proxy = (
            host for host in no_proxy.replace(' ', '').split(',') if host
        )

        if is_ipv4_address(parsed.hostname):
            for proxy_ip in no_proxy:
                if is_valid_cidr(proxy_ip):
                    if address_in_network(parsed.hostname, proxy_ip):
                        return True
                elif parsed.hostname == proxy_ip:
                    # If no_proxy ip was defined in plain IP notation instead of cidr notation &
                    # matches the IP of the index
                    return True
        else:
            host_with_port = parsed.hostname
            if parsed.port:
                host_with_port += ':{}'.format(parsed.port)

            for host in no_proxy:
                if parsed.hostname.endswith(host) or host_with_port.endswith(host):
                    # The URL does match something in no_proxy, so we don't want
                    # to apply the proxies on this URL.
                    return True

    with set_environ('no_proxy', no_proxy_arg):
        # parsed.hostname can be `None` in cases such as a file URI.
        try:
            bypass = proxy_bypass(parsed.hostname)
        except (TypeError, socket.gaierror):
            bypass = False

    if bypass:
        return True

    return False


def get_environ_proxies(url, no_proxy=None):
    """
    Return a dict of environment proxies.

    :rtype: dict
    """
    if should_bypass_proxies(url, no_proxy=no_proxy):
        return {}
    else:
        return getproxies()


def select_proxy(url, proxies):
    """Select a proxy for the url, if applicable.

    :param url: The url being for the request
    :param proxies: A dictionary of schemes or schemes and hosts to proxy URLs
    """
    proxies = proxies or {}
    urlparts = urlparse(url)
    if urlparts.hostname is None:
        return proxies.get(urlparts.scheme, proxies.get('all'))

    proxy_keys = [
        urlparts.scheme + '://' + urlparts.hostname,
        urlparts.scheme,
        'all://' + urlparts.hostname,
        'all',
    ]
    proxy = None
    for proxy_key in proxy_keys:
        if proxy_key in proxies:
            proxy = proxies[proxy_key]
            break

    return proxy


def default_user_agent(name="python-requests"):
    """
    Return a string representing the default user agent.

    :rtype: str
    """
    return '%s/%s' % (name, __version__)


def default_headers():
    """
    :rtype: requests.structures.CaseInsensitiveDict
    """
    return CaseInsensitiveDict({
        'User-Agent': default_user_agent(),
        'Accept-Encoding': ', '.join(('gzip', 'deflate')),
        'Accept': '*/*',
        'Connection': 'keep-alive',
    })


def parse_header_links(value):
    """Return a list of parsed link headers proxies.

    i.e. Link: <http:/.../front.jpeg>; rel=front; type="image/jpeg",<http://.../back.jpeg>; rel=back;type="image/jpeg"

    :rtype: list
    """

    links = []

    replace_chars = ' \'"'

    value = value.strip(replace_chars)
    if not value:
        return links

    for val in re.split(', *<', value):
        try:
            url, params = val.split(';', 1)
        except ValueError:
            url, params = val, ''

        link = {'url': url.strip('<> \'"')}

        for param in params.split(';'):
            try:
                key, value = param.split('=')
            except ValueError:
                break

            link[key.strip(replace_chars)] = value.strip(replace_chars)

        links.append(link)

    return links


# Null bytes; no need to recreate these on each call to guess_json_utf
_null = '\x00'.encode('ascii')  # encoding to ASCII for Python 3
_null2 = _null * 2
_null3 = _null * 3


def guess_json_utf(data):
    """
    :rtype: str
    """
    # JSON always starts with two ASCII characters, so detection is as
    # easy as counting the nulls and from their location and count
    # determine the encoding. Also detect a BOM, if present.
    sample = data[:4]
    if sample in (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE):
        return 'utf-32'     # BOM included
    if sample[:3] == codecs.BOM_UTF8:
        return 'utf-8-sig'  # BOM included, MS style (discouraged)
    if sample[:2] in (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE):
        return 'utf-16'     # BOM included
    nullcount = sample.count(_null)
    if nullcount == 0:
        return 'utf-8'
    if nullcount == 2:
        if sample[::2] == _null2:   # 1st and 3rd are null
            return 'utf-16-be'
        if sample[1::2] == _null2:  # 2nd and 4th are null
            return 'utf-16-le'
        # Did not detect 2 valid UTF-16 ascii-range characters
    if nullcount == 3:
        if sample[:3] == _null3:
            return 'utf-32-be'
        if sample[1:] == _null3:
            return 'utf-32-le'
        # Did not detect a valid UTF-32 ascii-range character
    return None


def prepend_scheme_if_needed(url, new_scheme):
    """Given a URL that may or may not have a scheme, prepend the given scheme.
    Does not replace a present scheme with the one provided as an argument.

    :rtype: str
    """
    scheme, netloc, path, params, query, fragment = urlparse(url, new_scheme)

    # urlparse is a finicky beast, and sometimes decides that there isn't a
    # netloc present. Assume that it's being over-cautious, and switch netloc
    # and path if urlparse decided there was no netloc.
    if not netloc:
        netloc, path = path, netloc

    return urlunparse((scheme, netloc, path, params, query, fragment))


def get_auth_from_url(url):
    """Given a url with authentication components, extract them into a tuple of
    username,password.

    :rtype: (str,str)
    """
    parsed = urlparse(url)

    try:
        auth = (unquote(parsed.username), unquote(parsed.password))
    except (AttributeError, TypeError):
        auth = ('', '')

    return auth


# Moved outside of function to avoid recompile every call
_CLEAN_HEADER_REGEX_BYTE = re.compile(b'^\\S[^\\r\\n]*$|^$')
_CLEAN_HEADER_REGEX_STR = re.compile(r'^\S[^\r\n]*$|^$')


def check_header_validity(header):
    """Verifies that header value is a string which doesn't contain
    leading whitespace or return characters. This prevents unintended
    header injection.

    :param header: tuple, in the format (name, value).
    """
    name, value = header

    if isinstance(value, bytes):
        pat = _CLEAN_HEADER_REGEX_BYTE
    else:
        pat = _CLEAN_HEADER_REGEX_STR
    try:
        if not pat.match(value):
            raise InvalidHeader("Invalid return character or leading space in header: %s" % name)
    except TypeError:
        raise InvalidHeader("Value for header {%s: %s} must be of type str or "
                            "bytes, not %s" % (name, value, type(value)))


def urldefragauth(url):
    """
    Given a url remove the fragment and the authentication part.

    :rtype: str
    """
    scheme, netloc, path, params, query, fragment = urlparse(url)

    # see func:`prepend_scheme_if_needed`
    if not netloc:
        netloc, path = path, netloc

    netloc = netloc.rsplit('@', 1)[-1]

    return urlunparse((scheme, netloc, path, params, query, ''))


def rewind_body(prepared_request):
    """Move file pointer back to its recorded starting position
    so it can be read again on redirect.
    """
    body_seek = getattr(prepared_request.body, 'seek', None)
    if body_seek is not None and isinstance(prepared_request._body_position, integer_types):
        try:
            body_seek(prepared_request._body_position)
        except (IOError, OSError):
            raise UnrewindableBodyError("An error occurred when rewinding request "
                                        "body for redirect.")
    else:
        raise UnrewindableBodyError("Unable to rewind request body for redirect.")



# -*- coding: utf-8 -*-
import sysactions
from sysactions import get_model

from logoff_user import logoffuser
from signout_user import sign_out_user

SIGNED_IN_USER = "SIGNED_IN_USER"
USECS_PER_SEC = 1000000


@sysactions.action
def common_logoff(pos_id, user_id=None, password=None):
    model = get_model(pos_id)
    working_mode = model.find("WorkingMode").get("usrCtrlType")

    if working_mode == "QSR":
        return logoffuser(pos_id, user_id, password)
    else:
        return sign_out_user(pos_id, user_id)



import sysactions

from xml.etree import cElementTree as eTree

from typing import (
    Dict,
    List,
    Union,
)

from _AposA8PrintLine import AposA8PrintLine
from _AposA8Align import AposA8Align
from _AposA8CommandType import AposA8CommandType
from _AposA8Constants import AposA8Constants

from common import PrinterCommon
from old_helper import F, get_sale_line_priced_items, round_half_away_from_zero

APOS_DASH = AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT)
APOS_NEWLINE = AposA8PrintLine(AposA8Align.CENTER, "", AposA8CommandType.PRINT_TEXT)
APOS_HASH = AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_HASH, AposA8CommandType.PRINT_TEXT)


def add_loyalty_custom_text(report, order, custom_text):
    # type: (list, eTree.Element, dict) -> None
    loyalty_user_name = order.find("CustomOrderProperties/OrderProperty[@key='CUSTOMER_NAME']")
    loyalty_user_cpf = order.find("CustomOrderProperties/OrderProperty[@key='LOYALTY_SYSTEM_USER_CPF']")

    if loyalty_user_name is not None and loyalty_user_cpf is not None:
        name_and_cpf = custom_text["name_cpf"].format(nome_cliente=loyalty_user_name.get("value"),
                                                      cpf_cliente=loyalty_user_cpf.get("value"))
        report.append(AposA8PrintLine(AposA8Align.CENTER, name_and_cpf, AposA8CommandType.PRINT_TEXT))
    elif loyalty_user_cpf is not None:
        cpf = custom_text["cpf"].format(cpf_cliente=loyalty_user_cpf.get("value"))
        report.append(AposA8PrintLine(AposA8Align.CENTER, cpf, AposA8CommandType.PRINT_TEXT))
    else:
        non_participating = custom_text["non_participating"]
        report.append(AposA8PrintLine(AposA8Align.CENTER, non_participating, AposA8CommandType.PRINT_TEXT))

    report.append(APOS_NEWLINE)
    report.append(APOS_DASH)
    report.append(APOS_NEWLINE)


def append_items_header(report):
    product_header = "COD     DESCR          QTD UN    VL UNIT   TOTAL"
    report.append(AposA8PrintLine(AposA8Align.CENTER, product_header, AposA8CommandType.PRINT_TEXT))


def append_and_count_sale_items(
    report,             # type: []
    order,              # type: eTree
    printer_common      # type: PrinterCommon
):
    # type: (...) -> int

    items_amount = 0
    sale_lines = get_sale_line_priced_items(order.findall("SaleLine"))

    for line in sale_lines:
        line_params = _get_line_params(line)
        item_price, multiplied_qty, name, part_code, level = line_params
        measure_unit = line.get("measureUnit") if "measureUnit" in line.attrib else "un"
        items_amount += 1
        quantity = F(multiplied_qty)
        item_value = round_half_away_from_zero(F(item_price) / quantity, 2)
        line = "%-7.7s %-14.14s%6.3f%-3.2s%9.2f" % (part_code, name, quantity, measure_unit, item_value)
        string = ("%-5.40s %7.02f" % (line, F(item_price)))

        report.append(AposA8PrintLine(AposA8Align.CENTER, string, AposA8CommandType.PRINT_TEXT))

    if float(order.get("tip", 0)) > 0:
        tips_lines = printer_common.get_tip_order(order, is_mobile=True)
        report.append(AposA8PrintLine(AposA8Align.CENTER, tips_lines, AposA8CommandType.PRINT_TEXT))

    return items_amount


def _get_line_params(
        line    # type: eTree
):
    # type: (...) -> Union[list, List[str]]

    return map(
                line.get,
                (
                    "itemPrice",
                    "multipliedQty",
                    "productName",
                    "partCode",
                    "level"
                )
        )

def append_tax_data(report, total_taxes):
    approx_value_text = "Valor aproximado dos tributos deste cupom"
    report.append(AposA8PrintLine(AposA8Align.LEFT, approx_value_text, AposA8CommandType.PRINT_TEXT))

    nacional_federal_value = total_taxes['nacionalfederal']['value']
    estadual_value = total_taxes['estadual']['value']

    total_tax_value = float(nacional_federal_value) + float(estadual_value)
    report.append(AposA8PrintLine(AposA8Align.RIGHT, "R$%7.2f" % total_tax_value, AposA8CommandType.PRINT_TEXT))

    nacional_federal_percent = total_taxes['nacionalfederal']['percent']
    estadual_percent = total_taxes['estadual']['percent']

    federal_tax = "Fed = R$%7.2f (%.2f%%), Est = R$%7.2f (%.2f%%)" % (
        nacional_federal_value,
        nacional_federal_percent,
        estadual_value,
        estadual_percent
    )
    report.append(AposA8PrintLine(AposA8Align.CENTER, federal_tax, AposA8CommandType.PRINT_TEXT))

    law_cite = "(Conforme Lei Fed. 12.741/2012)"
    report.append(AposA8PrintLine(AposA8Align.CENTER, law_cite, AposA8CommandType.PRINT_TEXT))

    report.append(APOS_DASH)


def append_coupon_fiscal_parameters(
    coupon_fiscal_parameters,   # type: Dict[str, bool]
    report,                     # type: List
    order,                      # type: eTree.Element
):
    # type: (...) -> None

    if not any(coupon_fiscal_parameters.values()):
        return

    if coupon_fiscal_parameters["operator"]:
        user_id = order.get("sessionId").split("user=")[1].split(",")[0]
        operator_name = eTree.XML(sysactions.get_user_information(user_id)).find(".//user").get("LongName")[:15]
        text = AposA8PrintLine(
            align=AposA8Align.CENTER,
            text="Operador: {}".format(operator_name),
            command_type=AposA8CommandType.PRINT_TEXT,
        )
        report.append(text)

    if coupon_fiscal_parameters["order_or_table"]:
        order_id = order.get("orderId")
        table_property = order.find("CustomOrderProperties/OrderProperty[@key='TABLE_NAME']")
        table_name = table_property.get("value") if table_property is not None else ''
        table_name = table_name[:15] + "..." if len(table_name) > 15 else table_name
        if table_name:
            report.append(
                AposA8PrintLine(
                    align=AposA8Align.CENTER,
                    text="Pedido/Mesa: {}/{}".format(order_id[-4:], table_name),
                    command_type=AposA8CommandType.PRINT_TEXT,
                )
            )
        else:
            report.append(
                AposA8PrintLine(
                    align=AposA8Align.CENTER,
                    text="Pedido: {}".format(order_id[-4:]),
                    command_type=AposA8CommandType.PRINT_TEXT,
                )
            )

    if coupon_fiscal_parameters["pos_id"]:
        pos_id = order.get("posId")
        report.append(
            AposA8PrintLine(
                align=AposA8Align.CENTER,
                text="POS ID: {}".format(pos_id),
                command_type=AposA8CommandType.PRINT_TEXT,
            )
        )

    report.append(
        AposA8PrintLine(
            align=AposA8Align.CENTER,
            text=AposA8Constants.SEPARATOR_HASH,
            command_type=AposA8CommandType.PRINT_TEXT
        )
    )



# -*- coding: utf-8 -*-
import base64
import time
from datetime import datetime, timedelta

from common import PrinterCommon
from helper import remove_accents, encode_cp860
from xml.etree import cElementTree as eTree
from xml.etree.cElementTree import Element

import constants
import printer_wrapper
import sysactions
from clichehelper import ClicheHelper
from dateutil import tz
from msgbus import FM_PARAM, TK_SYS_ACK, TK_PRN_PRINT, MBException, MBEasyContext
from old_helper import F, get_sale_line_priced_items, regex_search, round_half_away_from_zero
from persistence import Driver as DBDriver, AprDbdException
from typing import (
    Optional,
    Dict,
    List,
    Union,
    Any,
)

SATISFACTION_SURVEY_TEXT = None  # type: Optional[str]
PAYMENTS_DESCRIPTION = {}


class FiscalPrinter(object):
    def __init__(
        self,
        mb_context,                     # type: MBEasyContext
        print_customer_name,            # type: bool
        cliche_helper,                  # type: ClicheHelper
        coupon_fiscal_parameters=None,  # type: Optional[Dict[str, bool]]
    ):
        # type: (...) -> None

        self.mb_context = mb_context
        self.print_customer_name = print_customer_name
        self.cliche_helper = cliche_helper
        sysactions.mbcontext = mb_context
        self.coupon_fiscal_parameters = coupon_fiscal_parameters or {}

    def print_fiscal(self, pos_id, order, tenders, fiscal_request, fiscal_response, total_taxes, customer_name, url_sefaz=""):
        raise NotImplementedError()

    def print_cancel(self, pos_id, data):
        raise NotImplementedError()

    @staticmethod
    def _center(s, cols=constants.COLS):
        if not s:
            return s

        s = SatFiscalPrinter._cut(s)
        miss = cols - len(s)

        if miss == 0:
            return s

        left = miss / 2

        return (" " * left) + s

    @staticmethod
    def _cut(s):
        if len(s) > constants.COLS:
            # If there is a line-feed at the end, keep it
            return s[:constants.COLS] if (s[-1] != "\n") else (s[:(constants.COLS - 1)] + "\n")

        return s

    @staticmethod
    def get_satisfaction_survey_code(order):
        global SATISFACTION_SURVEY_TEXT

        satisfaction_survey_property = order.find(
            "CustomOrderProperties/OrderProperty[@key='SATISFACTION_SURVEY_CODE']")
        if satisfaction_survey_property is not None:
            satisfaction_survey_code = satisfaction_survey_property.get("value")
            if SATISFACTION_SURVEY_TEXT is None:
                SATISFACTION_SURVEY_TEXT = sysactions.get_storewide_config("Store.GuestSatisfactionSurvey.Text")
        else:
            satisfaction_survey_code = None
        return satisfaction_survey_code

    @staticmethod
    def get_payment_description(mb_context, tender):
        if not PAYMENTS_DESCRIPTION:
            _build_payments_description_cache(mb_context)

        tender_id = str(tender.type)
        try:
            tender_descr = PAYMENTS_DESCRIPTION[tender_id].get("tender_description")
            parent_tender_id = PAYMENTS_DESCRIPTION[tender_id].get("parent_tender_id")
            if parent_tender_id:
                parent_tender_descr = PAYMENTS_DESCRIPTION[parent_tender_id].get("tender_description")
                tender_descr = "{} {}".format(parent_tender_descr, tender_descr)
        except KeyError as _:
            return "Outros"

        return tender_descr

    def _manager_lines_size(self, line, max_size=35):
        fields = []
        report = ""

        while line != '':
            current_line = line[0:max_size]
            fields.append(report + current_line)
            line = line[len(current_line):]

        return fields


class DonationCouponPrinter(object):
    def __init__(self, cliche, nf_type):
        # type: (ClicheHelper, str) -> None
        self.nf_type = nf_type
        self.cliche = cliche

    def print_coupon(self, pos_id, order_id, customer_name, donation_value, donation_institution, donation_cnpj, donation_site):
        model = sysactions.get_model(pos_id)
        report = ""

        report += constants.LINESPACING_FUNCS[1]
        report += constants.TXT_FONT_B
        report += constants.TXT_ALIGN_CT
        for row in self.cliche.get_cliche():
            report += row + "\n"
        report += constants.SEPARATOR_DASH + "\n"
        report += constants.TXT_ALIGN_LT
        report += "%-44s   %s" % (datetime.now().strftime("%d/%m/%y"), datetime.now().strftime("%H:%M:%S")) + "\n"
        report += "NAO E DOCUMENTO FISCAL\n"
        report += constants.TXT_ALIGN_CT + constants.TXT_BOLD_ON + "COMPROVANTE NAO FISCAL" + constants.TXT_BOLD_OFF + "\n"
        report += constants.TXT_ALIGN_LT
        client_name = constants.TXT_BOLD_ON \
                      + (customer_name[:30] or _translate(model, "UNINFORMED")).upper() \
                      + "\n" \
                      + constants.TXT_BOLD_OFF
        report += "%-12s   %46s" % (_translate(model, "ORDER_NUMBER", "%s" % order_id),
                                    _translate(model, "CLIENT_NAME", "%s" % remove_accents(client_name))) + "\n"
        report += "\n" + constants.SEPARATOR_DASH + "\n"

        return report


class SatFiscalPrinter(FiscalPrinter):
    def __init__(
        self,
        mb_context,                 # type: MBEasyContext
        print_customer_name,        # type: bool
        cliche_helper,              # type: ClicheHelper
        customer_text_field,        # type: str
        print_order_identifier,     # type: bool
        printer_common,             # type: PrinterCommon
        coupon_fiscal_parameters,   # type: Dict[str, bool]
    ):
        # type: (...) -> None

        super(SatFiscalPrinter, self).__init__(
            mb_context,
            print_customer_name,
            cliche_helper,
            coupon_fiscal_parameters,
        )

        self.customer_text_field = customer_text_field
        self.print_order_identifier = print_order_identifier
        self.printer_common = printer_common

    def print_fiscal(self, pos_id, order, tenders, fiscal_request, fiscal_response, total_taxes, customer_name, url_sefaz=""):
        # type: (str, Element, list, Element, Element, dict, str, Optional[str]) -> None

        xml = base64.b64decode(fiscal_request.text.strip())
        parsed_xml = eTree.XML(xml)
        n_c_fe = parsed_xml.find("infCFe/ide/nCFe").text
        model = sysactions.get_model(pos_id)

        totem_print = sysactions.get_podtype(sysactions.get_model(pos_id)) == "TT"
        customer_name = remove_accents(customer_name)
        report = self._build_report(order, tenders, fiscal_response, n_c_fe, customer_name, totem_print, model)
        barcode1, barcode2 = printer_wrapper.barcode(fiscal_response.find("satkey").text.strip())

        report += barcode1
        report += barcode2
        report += constants.CTL_LF
        qr_code_text = fiscal_response.find("QRCode").text.strip()
        size = 8 if qr_code_text == "FakeSATQRCode" else 3
        fields = self._build_fields_footer(fiscal_response, customer_name, order)

        qr_image = printer_wrapper.qr(qr_code_text, size=size)
        report += printer_wrapper.insert_text_on_right_image(qr_image, fields)

        report += constants.CTL_LF
        report = _build_coupon_tax(report, total_taxes)
        report += _rollback_printer_default_font()

        return report

    def print_cancel(
            self,
            pos_id,     # type: str
            data        # type: str
    ):
        # type: (...) -> None

        fiscal_response = eTree.XML(data)
        xml = base64.b64decode(
                fiscal_response.find("XmlResponse").text.strip()
        )

        parsed_xml = eTree.XML(xml)
        n_c_fe = parsed_xml.find("infCFe/ide/nCFe").text
        total = parsed_xml.find("infCFe/total/vCFe").text
        report = self._cancel_report(
                fiscal_response.find("FiscalData"),
                n_c_fe,
                total
        )

        barcode1, barcode2 = printer_wrapper.barcode(
                fiscal_response.find("FiscalData/satkey").text.strip()
        )

        qr_image = printer_wrapper.qr(fiscal_response.find("FiscalData/QRCode").text.strip())
        qr_code = printer_wrapper.insert_text_on_right_image(qr_image, [])

        report = report.encode("utf-8") + barcode1
        report += barcode2
        report += "\n"
        report += qr_code
        report += constants.TXT_ALIGN_LT

        for _ in range(0, 2):
            self._print_with_retry(
                    pos_id,
                    report
            )

    def _print_with_retry(self, pos_id, report):
        msg = None
        printer = "printer%s" % pos_id
        try:
            msg = self.mb_context.MB_EasySendMessage(printer, TK_PRN_PRINT, FM_PARAM, report, 10000000)
        except MBException:
            time.sleep(0.5)
            msg = self.mb_context.MB_EasySendMessage(printer, TK_PRN_PRINT, FM_PARAM, report, 10000000)
        finally:
            if msg and msg.token != TK_SYS_ACK:
                raise MBException("Erro imprimindo: %d", msg.token)

    def _cancel_report(self, sat_fiscal_response, ncfe_number, total):
        report = unicode("", "utf-8")
        report += constants.LINESPACING_FUNCS[1]
        report += constants.TXT_FONT_B
        report += constants.TXT_ALIGN_CT

        for line in self.cliche_helper.get_cliche():
            report += line + "\n"
        report += constants.SEPARATOR_DASH
        report += "\n"
        report += constants.TXT_BOLD_ON + "EXTRATO No. %s" % ncfe_number
        report += "\n"
        report += "CUPOM FISCAL ELETRONICO - SAT"
        report += "\nCANCELAMENTO"
        report += "\n" + constants.TXT_ALIGN_CT + constants.SEPARATOR_DASH

        report += "\nDADOS DO CUPOM FISCAL ELETRONICO CANCELADO"
        cpf_cnpj = sat_fiscal_response.find("CustomerCPF").text
        report += "\nCPF/CNPJ do Consumidor: %s" % ("Nao informado" if not cpf_cnpj else cpf_cnpj)
        report += "\nTOTAL R$ %s" % total
        report = self._build_footer(report, sat_fiscal_response)

        return report

    def _build_report(self, order, tenders, fiscal_request, ncfe_number, customer_name, totem_print=False, model=None):
        # type: (Element, list, Element, dict, str, Optional[list], Optional[Element]) -> unicode

        report = _build_start_configs(self.cliche_helper, totem_print)
        report = self._build_header(ncfe_number, report, fiscal_request, totem_print)
        report = _common_fiscal_coupon(
            self.mb_context,
            customer_name,
            model,
            order,
            report,
            tenders,
            totem_print,
            self.print_customer_name,
            self.print_order_identifier,
            self.customer_text_field,
            self.printer_common,
            self.coupon_fiscal_parameters,
        )
        report = self._build_sat_key(report, fiscal_request)
        return report

    def _build_sat_key(self, report, sat_fiscal_response):
        report += constants.TXT_ALIGN_CT
        from printer_wrapper import insert_space
        sat_key = insert_space(sat_fiscal_response.find("satkey").text.strip(), 4)
        report += self._center(str(sat_key))
        report += "\n"
        return report

    def _build_footer(self, report, sat_fiscal_response):
        report += "\n" + constants.SEPARATOR_DASH + "\n" + constants.TXT_ALIGN_CT
        report += constants.TXT_BOLD_ON + "SAT No. %s" % sat_fiscal_response.find("satno").text.strip() + constants.TXT_BOLD_OFF
        report += "\n"
        report += sat_fiscal_response.find("datetime").text.strip()
        report += "\n"
        report += constants.TXT_ALIGN_CT
        from printer_wrapper import insert_space
        sat_key = insert_space(sat_fiscal_response.find("satkey").text.strip(), 4)
        report += self._center(str(sat_key))
        report += "\n"
        return report

    def _build_fields_footer(self, fiscal_response, customer_name, order):
        fields = []

        document = fiscal_response.find("CustomerCPF").text
        report = "Consumidor: {}".format("Não informado" if not document else document)
        fields.append(report)

        if self.print_customer_name and order.find(".//OrderProperty[@key='TABLE_ID']") is None:
            if customer_name:
                content = "Nome: {}".format(customer_name.upper().strip())
            else:
                content = "Pedido: {}".format(order.get('orderId')[-3:].upper().strip())
            field = "%s" % content
            fields += self._manager_lines_size(field, 40)

        sat_number = "SAT nº %s" % fiscal_response.find("satno").text.strip()
        report = sat_number
        fields.append(report)

        report = fiscal_response.find("datetime").text.strip()
        fields.append(report)
        fields.append("Consulte o QRCode pelo Aplicativo")
        fields.append("de olho na nota disponível na")
        fields.append("AppStore (Apple) e Play Store (Android)")
        return fields

    @staticmethod
    def _build_header(ncfe_number, report, fiscal_response, totem_print):
        report += constants.TXT_ALIGN_CT + constants.SEPARATOR_DASH
        report += "\n"
        report += constants.TXT_BOLD_ON + "EXTRATO No. %s" % ncfe_number
        report += "\n"
        report += u"CUPOM FISCAL ELETRÔNICO - SAT".encode('cp860') + constants.TXT_BOLD_OFF
        report += "\n" + constants.SEPARATOR_DASH
        cpf_cnpj = fiscal_response.find("CustomerCPF").text
        report += "\nCPF/CNPJ do Consumidor: %s" % ("Nao informado" if not cpf_cnpj else cpf_cnpj)
        report += "\n" + constants.TXT_ALIGN_LT + constants.SEPARATOR_DASH + constants.CTL_LF
        report += "\nCOD     DESCR              QTD UN    VL UNIT       TOTAL"
        report += "\n"
        return report


class NfceFiscalPrinter(FiscalPrinter):
    def __init__(
        self,
        mb_context,                 # type: MBEasyContext
        print_customer_name,        # type: bool
        cliche_helper,              # type: ClicheHelper
        qrcode_check_url,           # type: str
        versao_ws,                  # type: int
        customer_text_field,        # type: str
        print_order_identifier,     # type: bool
        printer_common,             # type: PrinterCommon
        coupon_fiscal_parameters,   # type: Dict[str, bool]
    ):
        # type: (...) -> None
        super(NfceFiscalPrinter, self).__init__(
            mb_context,
            print_customer_name,
            cliche_helper,
            coupon_fiscal_parameters,
        )
        self.qrcode_check_url = qrcode_check_url
        self.versao_ws = versao_ws
        self.customer_text_field = customer_text_field
        self.print_order_identifier = print_order_identifier
        self.printer_common = printer_common

    def print_fiscal(self, pos_id, order, tenders, fiscal_request, fiscal_response, total_taxes, customer_name, url_sefaz=""):
        # type: (str, Element, list, Element, Element, dict, str, str) -> unicode

        model = sysactions.get_model(pos_id)

        re_search = r"<[A-z0-9]*?:?{0}>(.*?)</([A-z0-9]*?):?{0}>"
        tp_emiss = regex_search(re_search, eTree.tostring(fiscal_request), key="tpEmis")

        totem_print = sysactions.get_podtype(model) == "TT"
        customer_name = remove_accents(customer_name)
        if tp_emiss == "9":
            report = self._build_report(order, tenders, fiscal_request, fiscal_response, total_taxes, "Via Estabelecimento", customer_name, self.qrcode_check_url, totem_print, model)
            report += "\n\n\n\n\n\n\n\n"
            report += constants.PAPER_PART_CUT
            report += self._build_report(order, tenders, fiscal_request, fiscal_response, total_taxes, "Via Consumidor", customer_name, self.qrcode_check_url, totem_print, model)
            return report

        report = self._build_report(order, tenders, fiscal_request, fiscal_response, total_taxes, "", customer_name, self.qrcode_check_url, totem_print, model)
        return report

    def print_cancel(self, pos_id, data):
        return ""

    def _build_report(self, order, tenders, fiscal_request, fiscal_response, total_taxes, via, customer_name, url_sefaz, totem_print=False, model=None):
        # type: (Element, list, Element, Element, dict, str, str, str, Optional[bool], Optional[Element]) -> str

        report = _build_start_configs(self.cliche_helper, totem_print)
        re_search, report, tp_emiss = self._build_header(fiscal_request, report)
        report = _common_fiscal_coupon(
            mb_context=self.mb_context,
            customer_name=customer_name,
            model=model,
            order=order,
            report=report,
            tenders=tenders,
            totem_print=totem_print,
            print_customer_name=self.print_customer_name,
            print_order_identifier=self.print_order_identifier,
            customer_text_field=self.customer_text_field,
            printer_common=self.printer_common,
            coupon_attrs=self.coupon_fiscal_parameters,
        )

        report = self._build_key_access(report, url_sefaz, fiscal_request)
        report += self._build_footer(fiscal_request, fiscal_response, re_search, tp_emiss, via)
        report = _build_coupon_tax(report, total_taxes)
        report += _rollback_printer_default_font()

        return report

    @staticmethod
    def _build_key_access(report, url_sefaz, fiscal_request):
        report += constants.TXT_ALIGN_CT + constants.TXT_BOLD_ON
        report += "\nConsulte pela Chave de Acesso em\n"
        report += constants.TXT_BOLD_OFF
        report += url_sefaz + "\n"
        re_search_attribute_id = r"<[A-z0-9]*?:?infNFe.*?Id=\"(.*?)\""
        chave = regex_search(re_search_attribute_id, eTree.tostring(fiscal_request), key="infNFe")[3:]
        for i in range(0, 44, 4):
            report += chave[i:i + 4] + " "
        report += "\n"

        return report

    def _build_footer(self, fiscal_request, fiscal_response, re_search, tp_emiss, via):

        customer_properties_search = r"<dest>[\s\S]*?<{0}>(.*?)</{0}>[\s\S]*?</dest>"
        cpf_cnpj = ""
        cpf = regex_search(customer_properties_search, eTree.tostring(fiscal_request), key="CPF")
        if cpf is not None:
            cpf_cnpj = cpf
        else:
            cnpj = regex_search(customer_properties_search, eTree.tostring(fiscal_request), key="CNPJ")
            if cnpj is not None:
                cpf_cnpj = cnpj

        # Consumidor
        fields = []
        report = "CONSUMIDOR %s" % ("NÃO INFORMADO" if not cpf_cnpj else ("- CPF/CNPJ " + cpf_cnpj))
        fields.append(report)

        numero = regex_search(re_search, eTree.tostring(fiscal_request), key="nNF")
        serie = regex_search(re_search, eTree.tostring(fiscal_request), key="serie")
        data_emissao_xml = regex_search(re_search, eTree.tostring(fiscal_request), key="dhEmi")
        data_emissao_str = _format_date(data_emissao_xml)

        report = "NFC-e Número %09d" % (int(numero))
        fields.append(report)
        report = "Série {:03d}  {}".format(int(serie), data_emissao_str)
        fields.append(report)

        tp_amb = regex_search(re_search, eTree.tostring(fiscal_request), key="tpAmb")
        if tp_emiss == "9":
            lines = ClicheHelper.manager_lines_size(via, 33)
            for line in lines:
                fields.append(line)

        def handle_datetime_str(date_string):
            from dateutil.parser import parse
            parsed_utc_date_string = parse(date_string)
            return parsed_utc_date_string.strftime("%d/%m/%YT%H:%M:%S.%f")[:-7]

        if fiscal_response is not None:
            protocol_number = regex_search(re_search, eTree.tostring(fiscal_response), key="nProt")
            if protocol_number is not None:

                text_information = "Protocolo de autorização: {}".format(protocol_number)
                lines_protocol = ClicheHelper.manager_lines_size(text_information, 45)
                for line in self.manager_lines(lines_protocol):
                    fields.append(line)

                protocol_date = regex_search(re_search, eTree.tostring(fiscal_response), key="dhRecbto")
                text_information = "Data de autorização: {}".format(handle_datetime_str(protocol_date).replace('T', ' '))
                lines_date = ClicheHelper.manager_lines_size(text_information, 45)
                for line in self.manager_lines(lines_date):
                    fields.append(line)

        if tp_amb == "2":
            report = "Emitida em Homologação"
            fields.append(report)
            report = "Sem Valor Fiscal"
            fields.append(report)
        if tp_emiss == "9":
            report = "Emitida em Contingência"
            fields.append(report)
            report = "Pendente de Autorização"
            fields.append(report)

        report += constants.CTL_LF
        url = regex_search(re_search, eTree.tostring(fiscal_request), key="qrCode")

        if fiscal_response and regex_search(re_search, eTree.tostring(fiscal_response), key="cMsg"):
            message_code = regex_search(re_search, eTree.tostring(fiscal_response), key="cMsg")
            message = regex_search(re_search, eTree.tostring(fiscal_response), key="xMsg")
            if message_code == '200' and message not in [None, '', ' ']:
                report += '\n' + message.replace('|', '\n') + '\n'

        qr_image = printer_wrapper.qr(url.strip(), size=4)
        report += printer_wrapper.insert_text_on_right_image(qr_image, fields)

        return report

    def manager_lines(self, lines):
        # type: (List[str]) -> str

        report = ""
        for index, line in enumerate(lines):
            if index == 0:
                index_separator = line.find(":")
                new_line = line[:index_separator] + line[index_separator:]
                report += new_line
                yield report
                continue
            yield line



    @staticmethod
    def _build_header(fiscal_request, report):
        report += constants.SEPARATOR_DASH
        report += "\n"
        report += "Documento Auxiliar da Nota Fiscal\n"
        report += "de Consumidor Eletronica\n"
        re_search = r"<[A-z0-9]*?:?{0}>(.*?)</([A-z0-9]*?):?{0}>"
        tp_emiss = regex_search(re_search, eTree.tostring(fiscal_request), key="tpEmis")
        if tp_emiss == "9":
            report += constants.SEPARATOR_DASH
            report += constants.TXT_BOLD_ON
            report += "EMITIDA EM CONTINGENCIA\n"
            report += constants.TXT_BOLD_OFF
            report += "Pendente de autorizacao\n"
        report += constants.TXT_ALIGN_LT + constants.SEPARATOR_DASH
        report += constants.TXT_BOLD_ON
        report += "COD     DESCR              QTD UN    VL UNIT       TOTAL\n"
        report += constants.TXT_BOLD_OFF
        return re_search, report, tp_emiss


def _common_fiscal_coupon(
    mb_context,                 # type: MBEasyContext
    customer_name,              # type: str
    model,                      # type: Element
    order,                      # type: Element
    report,                     # type: str
    tenders,                    # type: List
    totem_print,                # type: bool
    print_customer_name,        # type: bool
    print_order_identifier,     # type: bool
    customer_text_field,        # type: str
    printer_common,             # type: PrinterCommon
    coupon_attrs,               # type: Dict[str, bool]
):
    # type: (...) -> str

    sale_lines = get_sale_line_priced_items(order.findall("SaleLine"))
    items_amount, order_tip, report = _build_sale_lines(order, report, sale_lines, printer_common)
    report = _build_order_totals(mb_context, items_amount, order, order_tip, report, tenders)
    report = _build_customer_name(print_customer_name, customer_name, order, report)
    report = _build_satisfaction_survey_code(model, report, order)
    report = _build_order_identifier(
        report=report,
        order=order,
        coupon_attrs=coupon_attrs,
    )
    report = _get_delivery_info(order, report)
    report = _build_custom_text_field(customer_text_field, report)
    return report


def _build_sale_lines(order, report, sale_lines, printer_common):
    items_amount = 0
    for line in sale_lines:
        line_params = map(line.get, ("itemPrice", "multipliedQty", "productName", "partCode", "level"))
        item_price, multiplied_qty, name, part_code, level = line_params
        measure_unit = line.get("measureUnit") if "measureUnit" in line.attrib else "un"
        items_amount += 1
        quantity = F(multiplied_qty)
        item_value = round_half_away_from_zero(F(item_price) / quantity, 2)
        line = "%-7.7s %-18.18s%6.3f%2.2s%9.02f" % (part_code, name, quantity, measure_unit, item_value)
        string = ("%-46.46s %9.02f\n" % (line, F(item_price)))

        report += string.encode('cp860')
    order_tip = 0.0
    if order.get("tip") and float(order.get("tip")) > 0:
        order_tip = F(order.get("tip"))
        tips_lines = printer_common.get_tip_order(order)
        report += tips_lines
        items_amount += len(tips_lines.split("\n")) - 1

    return items_amount, order_tip, report


def _build_satisfaction_survey_code(model, report, order):
    satisfaction_survey_code = FiscalPrinter.get_satisfaction_survey_code(order)
    if satisfaction_survey_code is not None and model is not None:
        report += constants.TXT_ALIGN_CT
        report += constants.TXT_NORMAL
        ssc = satisfaction_survey_code
        satisfaction_survey_code = ssc[:3] + '-' + ssc[3:6] + '-' + ssc[6:9] + '-' + ssc[9:12] + '-' + ssc[12:15] + '-' + ssc[-1]
        report += SATISFACTION_SURVEY_TEXT + satisfaction_survey_code + "\n"
        report += constants.TXT_FONT_B
        report += constants.SEPARATOR_DASH + "\n"
    return report


def _build_customer_name(print_customer_name, customer_name, order, report):
    if print_customer_name and order.find(".//OrderProperty[@key='TABLE_ID']") is None:
        char_amount = 21
        content = (customer_name or '')[:(char_amount - 2)] or order.get('orderId')[-3:]
        report += constants.TXT_ALIGN_CT
        report += constants.TXT_BOLD_ON + encode_cp860("PAINEL DE RETIRADA - VOCÊ SERA CHAMADO POR:") + constants.TXT_BOLD_OFF
        report += "\n"
        report += constants.TXT_FONT_A + constants.TXT_INVERT_ON + constants.TXT_ALIGN_CT
        report += constants.TXT_SIZE + chr(constants.TXT_WIDTH[2] + constants.TXT_HEIGHT[2])

        total_whitespace = char_amount - len(content)
        half_whitespace = total_whitespace // 2
        odd_division = total_whitespace % 2 != 0
        report += " " * half_whitespace
        report += "%s" % content.upper()
        report += " " * (half_whitespace + (1 if odd_division else 0))

        report += constants.TXT_SIZE + chr(constants.TXT_WIDTH[1] + constants.TXT_HEIGHT[1])
        report += constants.TXT_FONT_B + constants.TXT_INVERT_OFF
        report += "\n" + constants.SEPARATOR_DASH + "\n"

    return report


def _get_initial_line(
):
    # type: (...) -> unicode

    line = constants.LINESPACING_FUNCS[1]
    line += constants.TXT_FONT_B
    line += constants.INITIALIZE_LINE
    return line


def _build_start_configs(cliche_helper, totem_print):
    report = ""

    report += _get_initial_line()
    report += constants.CHARCODE_PC860
    report += constants.TOTEM_LINE_SPACING
    report += constants.TXT_FONT_B

    report += cliche_helper.get_header_coupon()
    report += constants.TXT_ALIGN_CT
    return report


def _build_custom_text_field(customer_text_field, report):
    if customer_text_field != "":
        report += constants.TXT_ALIGN_CT
        customer_name = customer_text_field.replace("\\n", "\n")
        report += "\n" + "%s" % encode_cp860(customer_name)
        report += "\n" + constants.SEPARATOR_DASH + "\n"
    return report


def _build_order_identifier(
    report,                     # type: str
    order,                      # type: eTree.XML
    coupon_attrs,               # type: Dict
):
    # type: (...) -> unicode

    order_info = _get_order_info(
        order=order,
        coupon_attrs=coupon_attrs,
    )
    if order_info:
        report += order_info
        report += constants.SEPARATOR_DASH + "\n"

    return report


def _get_delivery_info(order, report):
    delivery_info = _get_delivery_partner_and_short_reference(order)
    if delivery_info:
        report += delivery_info
        report += "\n" + constants.SEPARATOR_DASH + "\n"

    return report


def _get_order_info(
    order,          # type: Dict
    coupon_attrs,   # type: Dict
):
    # type: (...) -> unicode

    report = ""
    if coupon_attrs["operator"]:
        user_id = order.get("sessionId").split("user=")[1].split(",")[0]
        operator_name = eTree.XML(sysactions.get_user_information(user_id)).find(".//user").get("LongName")[:15]
        report += "Operador: {} \n".format(operator_name)

    if coupon_attrs["order_or_table"]:
        order_id = order.get("orderId")
        table_property = order.find("CustomOrderProperties/OrderProperty[@key='TABLE_NAME']")
        table_name = table_property.get("value") if table_property is not None else ''
        table_name = table_name[:15] + "..." if len(table_name) > 15 else table_name
        if table_name:
            report += "Pedido/Mesa: {}/{} \n".format(order_id[-4:], table_name)
        else:
            report += "Pedido: {} \n".format(order_id[-4:])

    if coupon_attrs["pos_id"]:
        pos_id = order.get("posId")
        report += "POS ID: {} \n".format(pos_id)

    return report


def _build_coupon_tax(report, total_taxes):
    report += constants.TXT_FONT_B
    report += constants.TXT_ALIGN_CT
    total_tax_amount = float(total_taxes['nacionalfederal']['value']) + float(total_taxes['estadual']['value'])
    report += "%-44s   R$%7.2f\n" % ("Valor aproximado dos tributos deste cupom", total_tax_amount)
    report += "Fed = R$%7.2f (%.2f%%), Est = R$%7.2f (%.2f%%)" % (total_taxes['nacionalfederal']['value'],
                                                                  total_taxes['nacionalfederal']['percent'],
                                                                  total_taxes['estadual']['value'],
                                                                  total_taxes['estadual']['percent'])
    report += "\n(Conforme Lei Fed. 12.741/2012)"
    return report


def _rollback_printer_default_font():
    report = "{}{}{}".format(constants.HW_INIT, constants.TXT_NORMAL, constants.TXT_BOLD_OFF)
    report += "{}{}".format(constants.TXT_FONT_A, constants.TXT_ALIGN_LT)

    return report


def _build_order_totals(mb_context, items_amount, order, order_tip, report, tenders):
    discount_amount = float(order.get("discountAmount") or 0)
    total_amount = float(order.get("totalAmount"))
    tax_total = float(order.get("taxTotal"))
    order_voided = order.get("state") == "VOIDED"
    factor = -1 if order_voided else 1

    report += constants.SEPARATOR_DASH + "\n"
    report += "%-44s         %3d\n" % ("QTD. TOTAL DE ITENS", items_amount)
    report += "%-44s   R$%7.2f\n" % ("VALOR TOTAL", total_amount + tax_total + order_tip)

    if discount_amount:
        bonus_type = "DESCONTO" if discount_amount > 0 else "ACRESCIMO"
        report += "%-44s   R$%7.2f\n" % (bonus_type, abs(discount_amount))

    total_amount = total_amount + tax_total - discount_amount + order_tip
    report += "%-44s   R$%7.2f\n" % ("VALOR A PAGAR", total_amount)
    report += "\n%-44s  %s\n" % ("FORMA DE PAGAMENTO", "Valor Pago")

    for tender in tenders:
        tender_name = FiscalPrinter.get_payment_description(mb_context, tender)
        tender_descr = encode_cp860(tender_name)
        tender_amount = tender.amount
        report += "%-44s   R$%7.2f\n" % (tender_descr, tender_amount)

    change = tenders[len(tenders) - 1].change  # Get "Change" (sixth value of tender tuple) from last tender in tenders
    if change > 0:
        report += "%-44s   R$%7.2f\n" % ("Troco", change * factor)

    report += constants.SEPARATOR_DASH + "\n"
    return report


def _build_payments_description_cache(mb_context):
    global PAYMENTS_DESCRIPTION

    conn = DBDriver().open(mb_context)
    has_parent_tender_id_on_db = True
    try:
        result = conn.select("SELECT TenderId, TenderDescr, ParentTenderId FROM TenderType")
    except AprDbdException:
        has_parent_tender_id_on_db = False
        result = conn.select("SELECT TenderId, TenderDescr FROM TenderType")
    for row in result:
        tender_id = row.get_entry("TenderId")
        tender_descr = row.get_entry("TenderDescr")
        parent_tender_id = row.get_entry("ParentTenderId") if has_parent_tender_id_on_db else None
        PAYMENTS_DESCRIPTION[tender_id] = {"tender_description": tender_descr, "parent_tender_id": parent_tender_id}


def _format_date(data, mascara="%d/%m/%Y %H:%M:%S"):
    data_emissao_date_utc = data[:19]
    data_emissao_date_utc = datetime.strptime(data_emissao_date_utc, "%Y-%m-%dT%H:%M:%S")
    data_emissao_offset = int(data[20:22]) * 60 + int(data[23:])
    if data[19] == "+":
        data_emissao_offset *= -1
    data_emissao_date_utc = data_emissao_date_utc + timedelta(minutes=data_emissao_offset)
    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()

    data_emissao_date = data_emissao_date_utc.replace(tzinfo=from_zone)
    data_emissao_date = data_emissao_date.astimezone(to_zone)
    data_emissao_str = data_emissao_date.strftime(mascara)
    return data_emissao_str


def _translate(model, text, *param):
    return remove_accents(sysactions.translate_message(model, text, *param))


def _get_delivery_partner_and_short_reference(order):
    delivery_partner = order.find("CustomOrderProperties/OrderProperty[@key='PARTNER']")
    if delivery_partner is not None:
        short_reference = order.find("CustomOrderProperties/OrderProperty[@key='SHORT_REFERENCE']")
        if short_reference is not None:
            return "\n{}: {}".format(delivery_partner.get("value").upper(), short_reference.get("value").upper())

    return ""



from ._PrinterService import PrinterService
from ._FiscalService import FiscalService
from ._OrderService import OrderService
from ._ApiService import ApiService
from ._CouponService import CouponService
from ._ImagesService import ImagesService
from ._PaymentService import PaymentService
from ._ProductInfoService import ProductInfoService



# -*- coding: utf-8 -*-

import os

from typing import Optional

import cfgtools

from posot import OrderTaker
from sysactions import (
    show_messagebox,
    show_keyboard,
    StopAction,
    get_current_order,
    get_model
)

from utilfunctions import (
    validate_document,
    update_custom_properties,
    get_custom_property_value,
    input_document_pinpad_dialog
)
from model.customexception import RequiredDocumentException

document_required_value = None


def validate_required_document(
    pos_id,                     # type: str
    pos_ot,                     # type: OrderTaker
    total_gross,                # type: float
    amount,                     # type: str
    due_amount,                 # type: float
    ask_document_pinpad=False   # type: Optional[bool]
):
    # type: (...) -> None

    model = get_model(pos_id)
    required_value = _get_document_required_value()
    ask_for_doc = document_required_value != 0.0 and (float(total_gross) >= required_value)
    last_payment = amount == "" or (float(amount) >= float(due_amount))
    document = get_custom_property_value(pos_id, 'CUSTOMER_DOC')

    if ask_for_doc and last_payment and document is None:
        message = "$DOC_REQUIRED|%s" % "{0:.2f}".format(required_value)
        show_messagebox(pos_id, message, "$INFORMATION")

        document = get_customer_doc(
            pos_id,
            pos_ot,
            required=True,
            ask_document_pinpad=ask_document_pinpad
        )
        order_properties_dict = update_custom_properties(customer_doc=document)
        if len(order_properties_dict) > 0:
            order = get_current_order(model)
            pos_ot.setOrderCustomProperties(order_properties_dict, orderid=order.get("orderId"))


def get_customer_doc(
    pos_id,                      # type: str
    pos_ot,                      # type: OrderTaker
    default_doc="",              # type: Optional[str]
    required=False,              # type: bool
    ask_document_pinpad=False,   # type: Optional[bool]
    stop_action=True             # type: Optional[bool]
):
    # type: (...) -> str

    valid = False
    customer_doc = None
    while not valid:
        if ask_document_pinpad:
            customer_doc = input_document_pinpad_dialog(pos_id)

        if customer_doc is None:
            message = "$ONLY_NUMBERS"
            title = "$ENTER_CPF_CNPJ"
            customer_doc = show_keyboard(pos_id, message, title=title, mask="CPF_CNPJ", numpad=True, defvalue=default_doc)

        if customer_doc is None:
            if required:
                raise RequiredDocumentException()
            if stop_action:
                raise StopAction()
            return ""

        if customer_doc == "" and not required:
            customer_doc = ""
            break

        valid = validate_document(customer_doc)
        if not valid:
            customer_doc = None
            show_messagebox(pos_id, "$INVALID_CPF_CNPJ")

    fill_additional_info(pos_ot, "CPF", customer_doc)

    return customer_doc


def fill_additional_info(
        pos_ot,         # type: OrderTaker
        properties,     # type: str
        value           # type: str
):
    # type: (...) -> None

    if pos_ot.additionalInfo:
        pos_ot.additionalInfo += "|"
    else:
        pos_ot.additionalInfo = ""

    pos_ot.additionalInfo += "{0}={1}".format(properties, value)


def _get_document_required_value():
    # type: (...) -> float

    global document_required_value

    if document_required_value is None:
        config = cfgtools.read(os.environ["LOADERCFG"])
        document_required_value = float(config.find_value("Customizations.DocumentRequiredValue") or 0)

    return document_required_value



from _LoginSummaryInfo import LoginSummaryInfo
from _LoginSummaryDto import LoginSummaryDto
from _LoginSummaryHeaderDto import LoginSummaryHeaderDto
from _LoginSummaryBodyDto import LoginSummaryBodyDto

__all__ = ["LoginSummaryInfo",
           "LoginSummaryDto",
           "LoginSummaryHeaderDto",
           "LoginSummaryBodyDto"]



from ._CashReportFilterDto import CashReportFilterDto
from ._PeriodCashReportFilterDto import PeriodCashReportFilterDto

from ._CashReportDto import CashReportDto
from ._CashReportBodyDto import CashReportBodyDto
from ._CashReportBodyLineDto import CashReportBodyLineDto

__all__ = ["CashReportFilterDto",
           "PeriodCashReportFilterDto",
           "CashReportDto",
           "CashReportBodyDto",
           "CashReportBodyLineDto"
           ]



# -*- coding: utf-8 -*-

import logging

from msgbus import (
    TK_SLCTRL_OMGR_CLEARTENDERS,
    FM_PARAM,
    TK_SYS_ACK
)
from posot import OrderTakerException

from application.repository import OrderRepository
from actions import mb_context
from sysactions import (
    translate_message,
    get_model, send_message
)
from typing import (
    List,
    Optional, Tuple
)


class OrderInteractor(object):
    def __init__(self, pos_id):
        # type: (str) -> None

        self.logger = logging.getLogger("PosActions")
        self.order_repository = OrderRepository(mb_context)
        self.model = get_model(pos_id)
        self.pos_id = pos_id

    def get_orders_benefit(self, initial_date, end_date, operator_id):
        # type: (str, str, Optional[str]) -> List

        from msgbusboundary import MsgBusTableService as tableService
        try:
            all_coupons = translate_message(self.model, "ALL_COUPONS")
            pos_list = sorted(tableService().get_exclusive_pos_list())
            result = {'All': {'index': 0, 'id': None, 'name': all_coupons}}
            benefits = self.order_repository.get_orders_benefit(initial_date, end_date, operator_id, pos_list)
            index = 1
            for benefit in benefits:
                benefit_detail = benefit["benefit"]
                benefit_id = benefit_detail.get("id")
                if benefit_id in result.keys():
                    continue
                name = benefit_detail.get("userDescription").encode('utf-8')
                key = benefit_detail.get("name").encode('utf-8')
                data = {'index': index, 'id': benefit_id, 'name': name, 'key': key}
                result[benefit_id] = data
                index += 1

            return sorted(result.values(), key=lambda k: k['index'])
        except Exception as _:
            self.logger.exception("Error getting orders benefit")
            return []

    def clear_order_tender(
        self,
        order_tender_id,            # type: str
        order_id,                   # type: str
        blk_notify="0",             # type: str
        tender_id="",               # type: str
        keep_tender_records="0",    # type: str
    ):
        # type: (...) -> None

        params = (
            blk_notify,
            self.pos_id,
            tender_id,
            order_tender_id,
            keep_tender_records,
            order_id
        )
        self.send_order_message(params, TK_SLCTRL_OMGR_CLEARTENDERS)

    def send_order_message(
        self,
        params,     # type: Tuple
        token       # type: int
    ):
        params = "\0".join(str(p) for p in params)

        msg = send_message(
            "ORDERMGR%d" % int(self.pos_id),
            token,
            FM_PARAM,
            params,
            timeout=1000000 * 120
        )

        if msg.token != TK_SYS_ACK:
            params = msg.data.split('\0')
            self.logger.error("Error clear order tender: {} - {} - {}", params[0], params[0], params[1])

        return msg.data



from sysactions import get_used_service

from mbcontextmessagehandler import MbContextMessageBus
from messagebus import (
    Message,
    DataType,
)
from msgbus import TK_PRN_PRINT


class PrinterService(object):
    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
    ):
        # type: (...) -> None

        self.message_bus = message_bus

    def send_print_data(
        self,
        model,          # type: eTree
        response,       # type: str
    ):
        # type: (...) -> Message

        current_printer = get_used_service(
            model=model,
            service_type="printer",
        )

        message = Message(
            token=TK_PRN_PRINT,
            data_type=DataType.param,
            data=response,
        )

        return self.message_bus.send_message(
            component=current_printer,
            message=message,
        )



import json
from datetime import datetime

from sysactions import action
from typing import (
    Any,
    Dict,
    List,
    Union,
)

from .. import mb_context
from application.interactor import SectorOperatorInteractor


ALL_SECTORS = "$ALL_SECTORS"
SECTORS = []
OPERATORS_BY_SECTORS = {}


def get_sectors(
    pos_id,     # type: Any[Union[int, str]]
):
    # type: (...) -> List[str]

    global SECTORS
    if SECTORS:
        return SECTORS

    from msgbusboundary import MsgBusTableService as tableService

    sector_operator_interactor = SectorOperatorInteractor(table_service=tableService())
    sectors = sector_operator_interactor.get_sectors(pos_id=pos_id)
    SECTORS = json.loads(sectors or "[]")
    if ALL_SECTORS not in SECTORS:
        SECTORS.insert(0, ALL_SECTORS)

    return SECTORS


def get_operator_sector_info(
    pos_id,     # type: Any[Union[int, str]]
):
    # type: (...) -> Dict[str, List[int]]

    global OPERATORS_BY_SECTORS
    if OPERATORS_BY_SECTORS:
        return OPERATORS_BY_SECTORS

    from msgbusboundary import MsgBusTableService as tableService

    sector_operator_interactor = SectorOperatorInteractor(table_service=tableService())
    operators_by_sectors = sector_operator_interactor.get_operator_sector_info(pos_id=pos_id)
    OPERATORS_BY_SECTORS = json.loads(operators_by_sectors or "{}")

    return OPERATORS_BY_SECTORS


@action
def get_operator_sector_info_front(
    pos_id,     # type: str
):
    # type: (...) -> str

    operator_sector_info = get_operator_sector_info(pos_id=pos_id)
    operator_sector_info = json.dumps(operator_sector_info)

    return operator_sector_info


@action
def update_operator_sector_info(
    pos_id,         # type: Any[Union[int, str]]
    update_info,    # type: Dict
):
    # type: (...) -> None

    global OPERATORS_BY_SECTORS

    from msgbusboundary import MsgBusTableService as tableService

    sector_operator_interactor = SectorOperatorInteractor(table_service=tableService())
    updated_operator_sector_info = sector_operator_interactor.update_operator_sector_info(
        pos_id=pos_id,
        update_info=update_info,
    )
    OPERATORS_BY_SECTORS = json.loads(updated_operator_sector_info or "{}")

    mb_context.MB_EasyEvtSend(
        subject="SECTOR_INFO_MODIFIED",
        type="",
        xml=str(datetime.now())
    )
